import gzip
import logging
from pathlib import Path

import numpy as np
from astropy.table import Table


def write_results_as_hats(base_catalog_path, results, *, catalog_name=None, overwrite=False):
    """Write results to a HATS catalog.

    Parameters
    ----------
    base_catalog_path : str or Path
        The base path to the output hats directory.
    results : nested_pandas.NestedFrame
        The results to write, as a NestedFrame where each row is a sample.
    catalog_name : str, optional
        The name of the catalog to write. If None, the name will be derived from the
        base_catalog_path. Default: None
    overwrite : bool
        Whether to overwrite the output directory if it already exists.
        Default: False
    """
    base_catalog_path = Path(base_catalog_path)
    logging.debug(f"Writing results as HATS Catalog to {base_catalog_path}")

    # See if the (optional) LSDB package is installed.
    try:
        from lsdb import from_dataframe
    except ImportError as err:  # pragma: no cover
        raise ImportError(
            "The lsdb package is required to write results as HATS files. "
            "Please install it via 'pip install lsdb'."
        ) from err

    # Convert the results into an LSDB Catalog and output that. We just generate and output
    # the basic catalog (no margins or other extras).
    catalog = from_dataframe(results, ra_column="ra", dec_column="dec")
    catalog.write_catalog(
        base_catalog_path,
        catalog_name=catalog_name,
        as_collection=False,
        overwrite=overwrite,
    )


def read_numpy_data(file_path):
    """Read in a numpy array from different formats depending on the file extension.
    Automatically detects and handles files in .npy, .npz, .csv, .ecsv, and .txt
    formats.

    Parameters
    ----------
    file_path : str
        The path to the file to read.

    Returns
    -------
    data : numpy.ndarray
        The data read from the file.
    """
    file_path = Path(file_path)
    if not file_path.is_file():
        raise FileNotFoundError(f"File {file_path} not found.")

    # Load the data according to the format.
    if file_path.suffix == ".npy":
        data = np.load(file_path)
    elif file_path.suffix == ".npz":
        # For npz files, extract the first array
        data = np.load(file_path)["arr_0"]
    elif file_path.suffix in [".csv", ".ecsv"]:
        data = np.loadtxt(file_path, delimiter=",", comments="#")
    elif file_path.suffix in [".txt"]:
        data = np.loadtxt(file_path, comments="#")
    else:
        raise ValueError(f"Unsupported file format: {file_path.suffix}.")

    return data


def write_numpy_data(file_path, data):
    """Write a numpy array to a file in a format determined by the file extension.
    Automatically detects and handles files in .npy, .npz, .csv, .ecsv, and .txt
    formats.

    Parameters
    ----------
    file_path : str
        The path to the file to write.
    data : numpy.ndarray
        The data to write to the file.
    """
    file_path = Path(file_path)
    if file_path.suffix == ".npy":
        np.save(file_path, data)
    elif file_path.suffix == ".npz":
        np.savez_compressed(file_path, arr_0=data)
    elif file_path.suffix in [".csv", ".ecsv"]:
        np.savetxt(file_path, data, delimiter=",")
    elif file_path.suffix in [".txt", ".dat"]:
        np.savetxt(file_path, data)
    else:
        raise ValueError(f"Unsupported file format: {file_path.suffix}.")


def read_grid_data(input_file, format="ascii", validate=False):
    """Read 2-d grid data from a text, csv, ecsv, or fits file.

    Each line is of the form 'x0 x1 value' where x0 and x1 are the grid
    coordinates and value is the grid value. The rows should be sorted by
    increasing x0 and, within an x0 value, increasing x1.

    Parameters
    ----------
    input_file : str or file-like object
        The input data file.
    format : str
        The file format. Should be one of the formats supported by
        astropy Tables such as 'ascii', 'ascii.ecsv', or 'fits'.
        Default: 'ascii'
    validate : bool
        Perform additional validation on the input data.
        Default: False

    Returns
    -------
    x0 : numpy.ndarray
        A 1-d array with the values along the x-axis of the grid.
    x1 : numpy.ndarray
        A 1-d array with the values along the y-axis of the grid.
    values : numpy.ndarray
        A 2-d array with the values at each point in the grid with
        shape (len(x0), len(x1)).

    Raises
    ------
    ValueError if any data validation fails.
    """
    logging.debug(f"Loading file {input_file} (format={format})")
    if not Path(input_file).is_file():
        raise FileNotFoundError(f"File {input_file} not found.")

    data = Table.read(input_file, format=format, comment=r"\s*#")
    if len(data.colnames) != 3:
        raise ValueError(
            f"Incorrect format for grid data in {input_file} with format {format}. "
            f"Expected 3 columns but found {len(data.colnames)}."
        )
    x0_col = data.colnames[0]
    x1_col = data.colnames[1]
    v_col = data.colnames[2]

    # Get the values along the x0 and x1 dimensions.
    x0 = np.sort(np.unique(data[x0_col].data))
    x1 = np.sort(np.unique(data[x1_col].data))

    # Get the array of values.
    if len(data) != len(x0) * len(x1):
        raise ValueError(
            f"Incomplete data for {input_file} with format {format}. Expected "
            f"{len(x0) * len(x1)} entries but found {len(data)}."
        )

    # If we are validating, loop through the entire table and check that
    # the x0 and x1 values are in the expected order.
    if validate:
        counter = 0
        for i in range(len(x0)):
            for j in range(len(x1)):
                if data[x0_col][counter] != x0[i]:
                    raise ValueError(
                        f"Incorrect x0 ordering in {input_file} at line={counter}."
                        f"Expected {x0[i]} but found {data[x0_col][counter]}."
                    )
                if data[x1_col][counter] != x1[j]:
                    raise ValueError(
                        f"Incorrect x1 ordering in {input_file} at line={counter}. "
                        f"Expected {x1[j]} but found {data[x1_col][counter]}."
                    )
                counter += 1

    # Build the values matrix.
    values = data[v_col].data.reshape((len(x0), len(x1)))

    return x0, x1, values


def _read_lclib_data_from_open_file(input_file):
    """Read SNANA's lclib data from a text file.

    Parameters
    ----------
    input_file : file
        The input data file containing SNANA's lclib data.

    Returns
    -------
    curves : list of astropy.table.Table
        A list of Astropy Tables, each representing a light curve.
    """
    colnames = []
    curves = []
    meta = {}
    current_model = {}
    parnames = []
    in_doc_block = False

    for l_num, line in enumerate(input_file):
        # Strip out the trailing comment. Then skip lines that are either
        # empty or do not contain a key-value pair.
        line = line.split("#")[0].strip()
        if not line or ":" not in line:
            continue

        # Split the line into key and value.
        key, value = line.split(":", 1)
        value = value.strip()

        # Handle the keys corresponding to a documentation block.
        if key == "DOCUMENTATION":
            in_doc_block = True
        elif key == "DOCUMENTATION_END":
            in_doc_block = False
        if in_doc_block:
            # If we are in a documentation block, just continue to the next line.
            continue

        if key == "COMMENT":
            continue  # Skip comments.
        elif key == "FILTERS":
            # Create a list of data columns with time and each filter.
            colnames = ["time"]
            for c in value:
                colnames.append(c)
        elif key == "END_EVENT":
            curr_id = meta.get("id", "")
            if curr_id != value:
                raise ValueError(f"Event ID mismatch (line {l_num}): found {value}, expected {curr_id}.")

            # Save the table we have so far.
            curves.append(Table(current_model, meta=meta))
        elif key == "START_EVENT":
            if len(colnames) == 0:
                raise ValueError(f"Error on line= {l_num}: No filters defined.")

            # Start a new light curve, but resetting the lists of data from the columns.
            current_model["type"] = []  # Initialize the type list.
            for col in colnames:
                current_model[col] = []
            meta["id"] = value
        elif key == "S" or key == "T":
            # Save an observation or template to the current light curve.
            current_model["type"].append(key)  # Get the type from the key.

            # Get the time and magnitudes from the columns.
            col_vals = value.split()
            if len(col_vals) != len(colnames):
                raise ValueError(f"Expected {len(colnames)} values on line={l_num}: {col_vals}")
            for col_idx, col in enumerate(colnames):
                current_model[col].append(float(col_vals[col_idx]))
        elif key == "MODEL_PARNAMES" or key == "MODEL_PARAMETERS":
            parnames = value.split(",")
        elif key == "PARVAL":
            if "," in value:
                all_vals = value.split(",")
            else:
                all_vals = value.split()

            if len(all_vals) != len(parnames):
                raise ValueError(f"Expected {len(parnames)} parameter values on line={l_num}: {all_vals}")
            meta["PARVAL"] = {key: value for key, value in zip(parnames, all_vals, strict=False)}
        else:
            # Save everything else to the meta dictionary.
            meta[key] = value

    return curves


def read_lclib_data(input_file):
    """Read SNANA's LCLIB data from a text file.

    Parameters
    ----------
    input_file : str or Path
        The path to the SNANA LCLIB data file.

    Returns
    -------
    curves : list of astropy.table.Table
        A list of Astropy Tables, each representing a light curve.
    """
    input_file = Path(input_file)
    logging.debug(f"Loading SNANA LCLIB data from {input_file}")
    if not input_file.is_file():
        raise FileNotFoundError(f"File {input_file} not found.")

    # Use the file suffix to determine how to read the file.
    suffix = input_file.suffix.lower()
    if suffix in [".gz", ".gzip"]:
        # Open as a gzipped text file.
        with gzip.open(input_file, "rt") as file_ptr:
            curves = _read_lclib_data_from_open_file(file_ptr)
    elif suffix in [".dat", ".txt", ".text"]:
        # Try to open the file as a regular text file.
        with open(input_file, "r") as file_ptr:
            curves = _read_lclib_data_from_open_file(file_ptr)
    else:
        raise ValueError(f"Unsupported file format: {suffix}.")

    return curves



import numpy as np
import pandas as pd
from lightcurvelynx import _LIGHTCURVELYNX_TEST_DATA_DIR
from lightcurvelynx.obstable.roman_obstable import RomanObsTable


def make_random_apt_table():
    """Make random APT table for testing."""
    table = {
        "RA": 0.0,
        "DEC": 0.0,
        "PA": 0.0,
        "BANDPASS": "F129",
        "MA_TABLE_NUMBER": 1001,
        "DURATION": 100,
        "PLAN": 1,
        "PASS": 1,
        "SEGMENT": 1,
        "OBSERVATION": 1,
        "VISIT": 1,
        "EXPOSURE": 1,
    }
    return pd.DataFrame(table, index=[0])


def test_roman_obstable_init():
    """Test initializing RomanObsTable."""
    apt_table = make_random_apt_table()
    roman_obstable = RomanObsTable(apt_table)

    assert {"zp", "N_Eff_Pix", "zodi_countrate_min", "thermal_countrate", "exptime"}.issubset(
        roman_obstable.columns
    )


def test_calculate_skynoise():
    """Test calculate_skynoise function."""
    apt_table = make_random_apt_table()
    roman_obstable = RomanObsTable(
        apt_table,
        mat_table_path=_LIGHTCURVELYNX_TEST_DATA_DIR
        / "roman_characterization/roman_wfi_imaging_multiaccum_tables_with_exptime.csv",
    )

    zodi_scale = [1.0, 10.0]
    zodi_countrate = [0.0, 0.1]
    thermal_countrate = [0.1, 0.0]
    exptime = 10.0
    expected_skyvariance = [1.0, 10.0]
    skyvariance = roman_obstable.calculate_skynoise(exptime, zodi_scale, zodi_countrate, thermal_countrate)

    np.testing.assert_allclose(skyvariance, expected_skyvariance)


def test_noise_calculation():
    """Test that the noise calucation returns the right range."""
    apt_table = make_random_apt_table()
    roman_obstable = RomanObsTable(
        apt_table,
        mat_table_path=_LIGHTCURVELYNX_TEST_DATA_DIR
        / "roman_characterization/roman_wfi_imaging_multiaccum_tables_with_exptime.csv",
    )

    roman_obstable.survey_values["zodi_level"] = 2.0

    mag = np.array([24.5])
    expected_magerr = np.array([0.1])

    flux_nJy = np.power(10.0, -0.4 * (mag - 31.4))

    fluxerr_nJy = roman_obstable.bandflux_error_point_source(flux_nJy, 0)
    magerr = 1.086 * fluxerr_nJy / flux_nJy

    np.testing.assert_allclose(magerr, expected_magerr, rtol=0.2)



"""This is a file with helper functions for deriving different parameters from an ObsTable."""

import logging
from abc import ABC, abstractmethod

import numpy as np

from lightcurvelynx.astro_utils.mag_flux import mag2flux
from lightcurvelynx.astro_utils.zeropoint import (
    calculate_zp_from_maglim,
    sky_bg_adu_to_electrons,
)
from lightcurvelynx.consts import GAUSS_EFF_AREA2FWHM_SQ

logger = logging.getLogger(__name__)


class _ParamFormula:
    """Class to formulas for deriving parameters.

    Attributes
    ----------
    parameter : str
        The name of the parameter.
    inputs : list of str
        The names of the input parameters required for the formula.
    formula : function
        The function that computes the parameter from the inputs.
    """

    def __init__(self, parameter, inputs, formula):
        self.parameter = parameter
        self.inputs = inputs
        self.formula = formula

    def can_solve(self, params):
        """Check if the formula can be solved with the given parameters.

        Parameters
        ----------
        params : dict
            A dictionary mapping parameter names to their values.

        Returns
        -------
        bool
            True if the formula can be solved, False otherwise.
        """
        return all(param in params and params[param] is not None for param in self.inputs)

    def solve(self, params):
        """Call the formula with the given input parameters.

        Parameters
        ----------
        params : dict
            A dictionary mapping parameter names to their values.

        Returns
        -------
        The value of the derived parameter.
        """
        logger.debug(f"Deriving parameter {self.parameter} from {self.inputs}")
        return self.formula(**{k: params[k] for k in self.inputs})


class ParamDeriver(ABC):
    """Base class to derive parameters from an ObsTable using predefined formulas.

    Attributes
    ----------
    parameters : dict
        A dictionary mapping parameter names to their values. With None for parameters
        that do not have a valid value yet.
    formulas : list of _ParamFormula
        A list of _ParamFormula instances for deriving parameters.
    org_parameters : set of str
        A set of parameter names that are original parameters from the ObsTable.

    Parameters
    ----------
    parameter_list : iterable, optional
        A list of the parameters supported by this deriver. If None, no parameters are supported.
    """

    _registered_derivers = {}  # A mapping of class name to class object for subclasses.

    def __init__(self, parameter_list=None):
        if parameter_list is None:
            parameter_list = []
        self.parameters = {param: None for param in parameter_list}

        self.formulas = []
        self.org_parameters = set()
        self._init_formulas()

    def __init_subclass__(cls, **kwargs):
        # Register all subclasses in a dictionary mapping class name to the
        # class object, so we can programmatically create objects from the names.
        super().__init_subclass__(**kwargs)
        cls._registered_derivers[cls.__name__] = cls

    @classmethod
    def create_deriver(cls, deriver_name):
        """Create a deriver instance from its class name.

        Parameters
        ----------
        deriver_name : str
            The name of the deriver class to instantiate.

        Returns
        -------
        ParamDeriver
            An instance of the specified deriver class.

        Raises
        ------
        ValueError
            If the deriver name is not recognized.
        """
        if deriver_name not in cls._registered_derivers:
            raise ValueError(f"Unknown deriver name: {deriver_name}")
        return cls._registered_derivers[deriver_name]()

    def __str__(self):
        return (
            f"{self.__class__.__name__} with parameters: {list(self.parameters.keys())} "
            f"and {len(self.formulas)} formulas."
        )

    def add_formula(self, parameter, inputs, formula):
        """Add a new formula to the list of formulas.

        Parameters
        ----------
        parameter : str
            The name of the parameter to be derived.
        inputs : list of str
            The names of the input parameters required for the formula.
        formula : function
            The function that computes the parameter from the inputs.
        """
        # Make sure all the input and output parameters are in the parameters dict.
        if parameter not in self.parameters:
            raise KeyError(f"Parameter {parameter} is not a supported parameter.")
        for input_param in inputs:
            if input_param not in self.parameters:
                raise KeyError(f"Parameter {input_param} is not a supported parameter.")

        # Append the formula itself.
        self.formulas.append(_ParamFormula(parameter, inputs, formula))

    def init_from_obs_table(self, obs_table):
        """Initialize parameters from an ObsTable, retrieving values from either the
        columns or survey constants (survey_values) for every parameter it can find.

        Parameters
        ----------
        obs_table : ObsTable
            The observation table from which to initialize parameters.
        """
        # Get the filter row so we can unpack dictionaries if needed.
        filters = obs_table["filter"]

        # Load each parameter from the ObsTable if it exists and is valid.
        for param in self.parameters:
            if param in obs_table.columns:
                values = obs_table[param].to_numpy()
            elif param in obs_table.survey_values:
                values = obs_table.survey_values[param]
                if type(values) is dict:
                    values = np.array([values.get(band, None) for band in filters])
            else:
                values = None

            # Only use the values if they are valid (not None or NaN).
            if values is not None and not np.any(values == None):  # noqa: E711
                self.parameters[param] = values
                self.org_parameters.add(param)

    def derive_parameters(self, obs_table):
        """Derive parameters from an ObsTable using the predefined formulas, and
        update the ObsTable with the derived parameters.

        Parameters
        ----------
        obs_table : ObsTable
            The observation table from which to derive parameters.
        """
        logger.debug(f"Deriving parameters using {self.__class__.__name__}.")
        self.init_from_obs_table(obs_table)
        logger.debug(f"Starting parameter derivation with parameters: {self.parameters}")

        # Keep iterating through the formulas, trying to solve them, until we do a full iteration
        # through all of them without making any new progress.
        made_progress = True
        while made_progress:
            made_progress = False
            for formula in self.formulas:
                if self.parameters[formula.parameter] is None and formula.can_solve(self.parameters):
                    self.parameters[formula.parameter] = formula.solve(self.parameters)
                    made_progress = True
        logger.debug(f"After derivation, parameters = {self.parameters}")

        # Update the ObsTable with the derived parameters. Do not overrwrite anything already
        # there. Always prefer to add scalars to the survey values over adding entire columns.
        for param, value in self.parameters.items():
            if param not in self.org_parameters and value is not None:
                if np.isscalar(value):
                    obs_table.survey_values[param] = value
                elif np.max(np.abs(value - value[0])) < 1e-8:  # all values are effectively the same
                    obs_table.survey_values[param] = value[0]
                else:
                    obs_table.add_column(param, value)

    @abstractmethod
    def _init_formulas(self):
        """Initialize the formulas for deriving parameters. To be implemented by subclasses."""
        raise NotImplementedError("Subclasses must implement _init_formulas().")


class NoopParamDeriver(ParamDeriver):
    """A ParamDeriver that does nothing. Used as a placeholder."""

    def __init__(self):
        super().__init__(parameter_list=[])

    def _init_formulas(self):
        """Initialize the formulas for deriving parameters. To be implemented by subclasses."""
        pass


class FullParamDeriver(ParamDeriver):
    """Class to derive all supported parameters from an ObsTable using predefined formulas.

    Supported parameters (and their units) include:
    - adu_bias: Bias level in ADU
    - dark_current: Dark current in electrons / second / pixel
    - exptime: Exposure time in seconds
    - filter: Photometric filter (e.g., g, r, i)
    - fwhm_px: Full-width at half-maximum of the PSF in pixels
    - gain: CCD gain in electrons / ADU
    - maglim: Limiting magnitude (5-sigma) in mag
    - nexposure: Number of exposures per observation (unitless)
    - pixel_scale: Pixel scale in arcseconds per pixel
    - psf_footprint: Effective footprint of the PSF in pixels^2
    - read_noise: Read noise in electrons
    - seeing: Seeing in arcseconds
    - sky_bg_adu: Sky background in ADU / pixel
    - sky_bg_electrons: Sky background in electrons / pixel^2
    - skybrightness: Sky brightness in mag / arcsec^2
    - zp: Instrumental zero point (nJy per electron)
    - zp_per_band: Instrumental zero point per band (nJy per electron)

    Attributes
    ----------
    parameters : dict
        A dictionary mapping parameter names to their values. With None for parameters
        that do not have a valid value yet.
    formulas : list of _ParamFormula
        A list of _ParamFormula instances for deriving parameters.
    org_parameters : set of str
        A set of parameter names that are original parameters from the ObsTable.
    """

    def __init__(self):
        parameter_list = [
            "adu_bias",  # Bias level in ADU
            "dark_current",  # Dark current in electrons / second / pixel
            "exptime",  # Exposure time in seconds
            "filter",  # Photometric filter (e.g., g, r, i)
            "fwhm_px",  # Full-width at half-maximum of the PSF in pixels
            "gain",  # CCD gain in electrons / ADU
            "maglim",  # Limiting magnitude (5-sigma) in mag
            "nexposure",  # Number of exposures per observation (unitless)
            "pixel_scale",  # Pixel scale in arcseconds per pixel
            "psf_footprint",  # Effective footprint of the PSF in pixels
            "read_noise",  # Read noise in electrons
            "seeing",  # Seeing in arcseconds
            "sky_bg_adu",  # Sky background in ADU / pixel
            "sky_bg_electrons",  # Sky background in electrons / pixel^2
            "skybrightness",  # Sky brightness in mag / arcsec^2
            "zp",  # Instrumental zero point (nJy per electron)
            "zp_per_band",  # Instrumental zero point per band (nJy per electron)
        ]
        super().__init__(parameter_list=parameter_list)

    def _init_formulas(self):
        """Initialize the standard formulas for deriving parameters."""

        # Renaming transformations (makes a copy with a new name). The per-band information
        # has already been expanded out in init_from_obs_table().
        self.add_formula(parameter="zp", inputs=["zp_per_band"], formula=lambda zp_per_band: zp_per_band)

        # Formulas for deriving the psf_footprint (in pixels^2) and related information.
        self.add_formula(
            parameter="psf_footprint",
            inputs=["fwhm_px"],
            formula=lambda fwhm_px: GAUSS_EFF_AREA2FWHM_SQ * fwhm_px**2,
        )
        self.add_formula(
            parameter="fwhm_px",
            inputs=["pixel_scale", "seeing"],
            formula=lambda pixel_scale, seeing: seeing / pixel_scale,
        )
        self.add_formula(
            parameter="fwhm_px",
            inputs=["psf_footprint"],
            formula=lambda psf_footprint: np.sqrt(psf_footprint / GAUSS_EFF_AREA2FWHM_SQ),
        )
        self.add_formula(
            parameter="seeing",
            inputs=["pixel_scale", "fwhm_px"],
            formula=lambda pixel_scale, fwhm_px: fwhm_px * pixel_scale,
        )

        # Formulas for deriving the sky background (in electrons / pixel^2) and related information.
        self.add_formula(
            parameter="sky_bg_electrons",
            inputs=["skybrightness", "pixel_scale", "zp"],
            formula=lambda skybrightness, pixel_scale, zp: (mag2flux(skybrightness) * pixel_scale**2 / zp),
        )
        self.add_formula(
            parameter="sky_bg_electrons",
            inputs=["sky_bg_adu", "gain"],
            formula=sky_bg_adu_to_electrons,
        )
        self.add_formula(
            parameter="sky_bg_adu",
            inputs=["sky_bg_electrons", "gain"],
            formula=lambda sky_electrons, gain: sky_electrons / gain,
        )

        # Formulas for deriving the zero point (in nJy per electron) and related information.
        self.add_formula(
            parameter="zp",
            inputs=[
                "maglim",
                "sky_bg_electrons",
                "fwhm_px",
                "read_noise",
                "dark_current",
                "exptime",
                "nexposure",
            ],
            formula=calculate_zp_from_maglim,
        )


class FiveSigmaDepthDeriver(ParamDeriver):
    """Class to derive the noise parameters from only the five-sigma depth information.

    Supported parameters (and their units) include:
    - bandflux_error: The error associated with the computed bandflux.
    - bandflux_ref: The total flux that would be transmitted through the given bandfilter.
    - five_sigma_depth: Five-sigma depth in magnitudes

    Attributes
    ----------
    parameters : dict
        A dictionary mapping parameter names to their values. With None for parameters
        that do not have a valid value yet.
    formulas : list of _ParamFormula
        A list of _ParamFormula instances for deriving parameters.
    org_parameters : set of str
        A set of parameter names that are original parameters from the ObsTable.
    """

    def __init__(self):
        parameter_list = [
            "bandflux_error",  # The error associated with the computed bandflux.
            "bandflux_ref",  # The total flux that would be transmitted through the given bandfilter.
            "five_sigma_depth",  # Five-sigma depth in magnitudes
        ]
        super().__init__(parameter_list=parameter_list)

    @staticmethod
    def _bandflux_error_from_five_sigma_depth(five_sigma_depth, bandflux_ref):
        """Derive the bandflux error from the five-sigma depth and bandflux reference.

        Based on redback's bandflux_error_from_limiting_mag()
        https://github.com/nikhil-sarin/redback

        Parameters
        ----------
        five_sigma_depth : float or np.ndarray
            The five-sigma depth in magnitudes.
        bandflux_ref : float or np.ndarray
            The reference bandflux.

        Returns
        -------
        bandflux_error : float or np.ndarray
            The error associated with the computed bandflux.
        """
        flux_five_sigma = bandflux_ref * np.power(10.0, -0.4 * five_sigma_depth)
        bandflux_error = flux_five_sigma / 5.0
        return bandflux_error

    def _init_formulas(self):
        self.add_formula(
            parameter="bandflux_error",
            inputs=["five_sigma_depth", "bandflux_ref"],
            formula=FiveSigmaDepthDeriver._bandflux_error_from_five_sigma_depth,
        )



import numpy as np
from lightcurvelynx.astro_utils.mag_flux import flux2mag
from lightcurvelynx.astro_utils.zeropoint import (
    calculate_zp_from_maglim,
    flux_electron_zeropoint,
    magnitude_electron_zeropoint,
    sky_bg_adu_to_electrons,
)
from lightcurvelynx.obstable.opsim import LSSTCAM_PIXEL_SCALE, _lsstcam_dark_current, _lsstcam_readout_noise
from scipy.optimize import fsolve


def test_magnitude_electron_zeropoint():
    """Test that instrumental zeropoints are correct"""
    # Reproducing magnitude corresponding to S/N=5 from
    # https://smtn-002.lsst.io/v/OPSIM-1171/index.html
    fwhm_eff = {
        "u": 0.92,
        "g": 0.87,
        "r": 0.83,
        "i": 0.80,
        "z": 0.78,
        "y": 0.76,
    }
    fwhm_eff_getter = np.vectorize(fwhm_eff.get)
    sky_brightness = {
        "u": 23.05,
        "g": 22.25,
        "r": 21.20,
        "i": 20.46,
        "z": 19.61,
        "y": 18.60,
    }
    sky_brightness_getter = np.vectorize(sky_brightness.get)

    # Use the LSST extinction coefficients and zeropoints per second at zenith.
    extinction_coeff = {
        "u": -0.458,
        "g": -0.208,
        "r": -0.122,
        "i": -0.074,
        "z": -0.057,
        "y": -0.095,
    }
    zp_per_sec_zenith = {
        "u": 26.524,
        "g": 28.508,
        "r": 28.361,
        "i": 28.171,
        "z": 27.782,
        "y": 26.818,
    }

    bands = list(fwhm_eff.keys())
    exptime = 30
    airmass = 1
    s2n = 5
    zp = magnitude_electron_zeropoint(
        filter=bands,
        airmass=airmass,
        exptime=exptime,
        instr_zp_mag=zp_per_sec_zenith,
        ext_coeff=extinction_coeff,
    )
    sky_count_per_arcsec_sq = np.power(10.0, -0.4 * (sky_brightness_getter(bands) - zp))
    readout_per_arcsec_sq = _lsstcam_readout_noise**2 / LSSTCAM_PIXEL_SCALE**2
    dark_per_arcsec_sq = _lsstcam_dark_current * exptime / LSSTCAM_PIXEL_SCALE**2
    count_per_arcsec_sq = sky_count_per_arcsec_sq + readout_per_arcsec_sq + dark_per_arcsec_sq

    area = 2.266 * fwhm_eff_getter(bands) ** 2  # effective seeing area in arcsec^2
    n_background = count_per_arcsec_sq * area
    # sky-dominated regime would be n_signal = s2n * np.sqrt(n_background)
    n_signal = 0.5 * (s2n**2 + np.sqrt(s2n**4 + 4 * s2n**2 * n_background))
    mag_signal = zp - 2.5 * np.log10(n_signal)

    m5_desired = {
        "u": 23.70,
        "g": 24.97,
        "r": 24.52,
        "i": 24.13,
        "z": 23.56,
        "y": 22.55,
    }
    assert list(m5_desired) == bands
    m5_desired_getter = np.vectorize(m5_desired.get)

    np.testing.assert_allclose(mag_signal, m5_desired_getter(bands), atol=0.1)

    # Test that the same result is obtained if we pass in the full arrays.
    # The arrays must be in the same order as the bands.
    extinction_coeff_array = np.array([extinction_coeff[b] for b in bands])
    zp_per_sec_zenith_array = np.array([zp_per_sec_zenith[b] for b in bands])
    zp2 = magnitude_electron_zeropoint(
        filter=bands,
        airmass=airmass,
        exptime=exptime,
        instr_zp_mag=zp_per_sec_zenith_array,
        ext_coeff=extinction_coeff_array,
    )
    np.testing.assert_allclose(zp, zp2, rtol=1e-10)


def test_flux_electron_zeropoint():
    """Test that bandflux zeropoints are correct"""
    extinction_coeff = {
        "u": -0.458,
        "g": -0.208,
        "r": -0.122,
        "i": -0.074,
        "z": -0.057,
        "y": -0.095,
    }
    zp_per_sec_zenith = {
        "u": 26.524,
        "g": 28.508,
        "r": 28.361,
        "i": 28.171,
        "z": 27.782,
        "y": 26.818,
    }

    # Here we just check that magnitude-flux conversion is correct
    airmass = np.array([1, 1.5, 2]).reshape(-1, 1, 1)
    exptime = np.array([30, 38, 45]).reshape(1, -1, 1)
    bands = ["u", "g", "r", "i", "z", "y"]
    mag = magnitude_electron_zeropoint(
        filter=bands,
        airmass=airmass,
        exptime=exptime,
        instr_zp_mag=zp_per_sec_zenith,
        ext_coeff=extinction_coeff,
    )
    flux = flux_electron_zeropoint(
        filter=bands,
        airmass=airmass,
        exptime=exptime,
        instr_zp_mag=zp_per_sec_zenith,
        ext_coeff=extinction_coeff,
    )
    np.testing.assert_allclose(mag, flux2mag(flux), rtol=1e-10)

    # Test that the same result is obtained if we pass in the full arrays.
    # The arrays must be in the same order as the bands.
    extinction_coeff_array = np.array([extinction_coeff[b] for b in bands])
    zp_per_sec_zenith_array = np.array([zp_per_sec_zenith[b] for b in bands])
    flux2 = flux_electron_zeropoint(
        filter=bands,
        airmass=airmass,
        exptime=exptime,
        instr_zp_mag=zp_per_sec_zenith_array,
        ext_coeff=extinction_coeff_array,
    )
    np.testing.assert_allclose(flux, flux2, rtol=1e-10)


def _fluxeq(
    flux, sky=None, gain=None, read_noise=None, nexposure=1, fwhm_px=None, dark_current=None, exptime=None
):
    """Define the equation to solve for flux at 5-sigma limit"""

    npix = 2.266 * fwhm_px**2  # = 4 * pi * sigma**2 = pi/2/ln2 * FWHM**2
    y = (
        flux**2
        - 25 * flux
        - 25
        * (sky * npix * gain + read_noise**2 * nexposure * npix + dark_current * npix * exptime * nexposure)
    )
    return y


def test_zp_from_maglim():
    """Test the calculation with numerical solution to the maglim function."""
    maglim = 19.0
    sky = 100.0
    fwhm = 1.5
    gain = 1.0
    read_noise = 10.0
    dark_current = 0.1
    exptime = 30.0

    root = fsolve(
        lambda x: _fluxeq(
            x,
            sky=sky,
            gain=gain,
            read_noise=read_noise,
            fwhm_px=fwhm,
            dark_current=dark_current,
            exptime=exptime,
        ),
        [100],
    )
    zp_expected = 2.5 * np.log10(root) + maglim

    zp_cal = calculate_zp_from_maglim(
        maglim=maglim,
        sky_bg_electrons=sky_bg_adu_to_electrons(sky, gain),
        fwhm_px=fwhm,
        read_noise=read_noise,
        dark_current=dark_current,
        exptime=exptime,
        nexposure=1,
    )

    assert np.isclose(zp_cal, zp_expected)



"""Samplers used for testing that produces precomputed results. These
can be used in testing to produce known results or to use data previously
sampled from another method (such as pzflow).
"""

import numpy as np
import pandas as pd
from astropy.table import Table

from lightcurvelynx.base_models import FunctionNode
from lightcurvelynx.math_nodes.np_random import NumpyRandomFunc


class BinarySampler(NumpyRandomFunc):
    """A FunctionNode that randomly returns True or False according
    to a given probability. This function is particularly useful in
    probabilistically applying effects or making decisions in the
    simulation.

    Attributes
    ----------
    probability : float
        The probability of returning True.
    """

    def __init__(self, probability, seed=None, **kwargs):
        if probability < 0 or probability > 1:
            raise ValueError("Probability must be between 0 and 1.")
        self.probability = probability

        super().__init__("uniform", seed=seed, **kwargs)

    def compute(self, graph_state, rng_info=None, **kwargs):
        """Return the given values.

        Parameters
        ----------
        graph_state : GraphState
            An object mapping graph parameters to their values. This object is modified
            in place as it is sampled.
        rng_info : numpy.random._generator.Generator, optional
            A given numpy random number generator to use for this computation. If not
            provided, the function uses the node's random number generator.
        **kwargs : dict, optional
            Additional function arguments.

        Returns
        -------
        results : any
            The result of the computation. This return value is provided so that testing
            functions can easily access the results.
        """
        rng = rng_info if rng_info is not None else self._rng

        if graph_state.num_samples == 1:
            results = rng.random() < self.probability
        else:
            results = rng.random(graph_state.num_samples) < self.probability
        self._save_results(results, graph_state)

        return results


class GivenValueList(FunctionNode):
    """A FunctionNode that returns given results for a single parameter
    in the order in which they are provided.

    Note
    ----
    This node does not support parallel sampling. It will return the same
    sequence of values for for each shard.

    Attributes
    ----------
    values : float, list, or numpy.ndarray
        The values to return.
    next_ind : int
        The index of the next value.
    """

    def __init__(self, values, **kwargs):
        self.values = np.asarray(values)
        if len(values) == 0:
            raise ValueError("No values provided for GivenValueList")
        self.next_ind = 0

        super().__init__(self._non_func, **kwargs)

    def reset(self):
        """Reset the next index to use."""
        self.next_ind = 0

    def compute(self, graph_state, rng_info=None, **kwargs):
        """Return the given values.

        Parameters
        ----------
        graph_state : GraphState
            An object mapping graph parameters to their values. This object is modified
            in place as it is sampled.
        rng_info : numpy.random._generator.Generator, optional
            Unused in this function, but included to provide consistency with other
            compute functions.
        **kwargs : dict, optional
            Additional function arguments.

        Returns
        -------
        results : any
            The result of the computation. This return value is provided so that testing
            functions can easily access the results.
        """
        if graph_state.num_samples == 1:
            if self.next_ind >= len(self.values):
                raise IndexError()

            results = self.values[self.next_ind]
            self.next_ind += 1
        else:
            end_ind = self.next_ind + graph_state.num_samples
            if end_ind > len(self.values):
                raise IndexError()

            results = self.values[self.next_ind : end_ind]
            self.next_ind = end_ind

        # Save and return the results.
        self._save_results(results, graph_state)
        return results


class GivenValueSampler(NumpyRandomFunc):
    """A FunctionNode that returns randomly selected items from a given list
    with replacement.

    Attributes
    ----------
    values : int, list, or numpy.ndarray
        The values to select from. If an integer is provided, it is treated as a range
        from 0 to that value - 1.
    _num_values : int
        The number of values that can be sampled.
    _weights : numpy.ndarray, optional
        The weights for each value, if provided. If None, all values are equally likely.
    """

    def __init__(self, values, weights=None, seed=None, **kwargs):
        if isinstance(values, int):
            values = np.arange(values)
        self.values = np.asarray(values)

        self._num_values = len(values)
        if self._num_values == 0:
            raise ValueError("No values provided for NumpySamplerNode")

        # Compute the normalized weights for each value.
        if weights is not None:
            self._weights = np.asarray(weights)
            if len(self._weights) != self._num_values:
                raise ValueError("Weights must match the number of values provided.")
            self._weights /= np.sum(self._weights)
        else:
            self._weights = None

        super().__init__("uniform", seed=seed, **kwargs)

    def compute(self, graph_state, rng_info=None, **kwargs):
        """Return the given values.

        Parameters
        ----------
        graph_state : GraphState
            An object mapping graph parameters to their values. This object is modified
            in place as it is sampled.
        rng_info : numpy.random._generator.Generator, optional
            A given numpy random number generator to use for this computation. If not
            provided, the function uses the node's random number generator.
        **kwargs : dict, optional
            Additional function arguments.

        Returns
        -------
        results : any
            The result of the computation. This return value is provided so that testing
            functions can easily access the results.
        """
        rng = rng_info if rng_info is not None else self._rng

        if graph_state.num_samples == 1:
            inds = rng.choice(self._num_values, p=self._weights)
        else:
            inds = rng.choice(self._num_values, size=graph_state.num_samples, p=self._weights)
        results = self.values[inds]
        self._save_results(results, graph_state)

        return results


class GivenValueSelector(FunctionNode):
    """A FunctionNode that selects a single value from a list of parameters.

    Parameters
    ----------
    values : float, list, or numpy.ndarray
        The values that can be selected.
    index : parameter
        The parameter that selects which value to return. This should return an
        integer index corresponding to the position in `values`.
    **kwargs : dict, optional
        Any additional keyword arguments.
    """

    def __init__(self, values, index, **kwargs):
        # The index parameter will automatically be added as input by the FunctionNode constructor.
        super().__init__(self._select, index=index, **kwargs)
        self.values = np.asarray(values)
        if len(values) == 0:
            raise ValueError("No values provided for GivenValueList")

    def _select(self, index):
        """Select the value at the given index."""
        if np.any(index < 0) or np.any(index >= len(self.values)):
            raise IndexError(f"Index {index} out of bounds for values of length {len(self.values)}")
        return self.values[index]


class TableSampler(NumpyRandomFunc):
    """A FunctionNode that returns values from a table-like data,
    including a Pandas DataFrame or AstroPy Table. The results returned
    can be in-order (for testing) or randomly selected with replacement.

    Note
    ----
    This node does not support "in order" parallel sampling. It will
    return the same sequence of values for each shard.

    Parameters
    ----------
    data : pandas.DataFrame, astropy.table.Table, or dict
        The object containing the data to sample.
    in_order : bool
        Return the given data in order of the rows (True). If False, performs
        random sampling with replacement. Default: False

    Attributes
    ----------
    data : astropy.table.Table
        The object containing the data to sample.
    in_order : bool
        Return the given data in order of the rows (True). If False, performs
        random sampling with replacement. Default: False
    next_ind : int
        The next index to sample for in order sampling.
    num_values : int
        The total number of items from which to draw the data.
    """

    def __init__(self, data, in_order=False, **kwargs):
        self.next_ind = 0
        self.in_order = in_order

        if isinstance(data, dict):
            self.data = Table(data)
        elif isinstance(data, Table):
            self.data = data.copy()
        elif isinstance(data, pd.DataFrame):
            self.data = Table.from_pandas(data)
        else:
            raise TypeError("Unsupported data type for TableSampler.")

        # Check there are some rows.
        self._num_values = len(self.data)
        if self._num_values == 0:
            raise ValueError("No data provided to TableSampler.")

        # Add each of the flow's data columns as an output parameter.
        super().__init__("uniform", outputs=self.data.colnames, **kwargs)

    def reset(self):
        """Reset the next index to use. Only used for in-order sampling."""
        self.next_ind = 0

    def compute(self, graph_state, rng_info=None, **kwargs):
        """Return the given values.

        Parameters
        ----------
        graph_state : GraphState
            An object mapping graph parameters to their values. This object is modified
            in place as it is sampled.
        rng_info : numpy.random._generator.Generator, optional
            A given numpy random number generator to use for this computation. If not
            provided, the function uses the node's random number generator.
        **kwargs : dict, optional
            Additional function arguments.

        Returns
        -------
        results : any
            The result of the computation. This return value is provided so that testing
            functions can easily access the results.
        """
        # Compute the indices to sample.
        if self.in_order:
            # Check that we have enough points left to sample.
            end_index = self.next_ind + graph_state.num_samples
            if end_index > len(self.data):
                raise IndexError()

            sample_inds = np.arange(self.next_ind, end_index)
            self.next_ind = end_index
        else:
            rng = rng_info if rng_info is not None else self._rng
            sample_inds = rng.integers(0, self._num_values, size=graph_state.num_samples)

        # Parse out each column into a separate parameter with the column name as its name.
        results = []
        for attr_name in self.outputs:
            attr_values = np.asarray(self.data[attr_name][sample_inds])
            if graph_state.num_samples == 1:
                results.append(attr_values[0])
            else:
                results.append(attr_values)

        # Save and return the results.
        self._save_results(results, graph_state)
        return results



import numpy as np
from astropy import constants as const


def bulk_convert_matrix(mat, in_units, out_units):
    """Bulk convert a numpy matrix from one set of units to another.

    Parameters
    ----------
    mat : numpy.ndarray
        The numpy matrix of values to convert.
    in_units : astropy.units
        The current (input) units of the matrix.
    out_units : astropy.units
        The desired (output) units of the matrix.

    Returns
    -------
    output : numpy.ndarray
        The converted numpy matrix of values.
    """
    multipler = (1.0 * in_units).to_value(out_units)
    return multipler * mat


def flam_to_fnu(flux_flam, wavelengths, *, wave_unit, flam_unit, fnu_unit):
    """Covert flux from f_lambda unit to f_nu unit.

    Parameters
    ----------
    flux_flam : list or numpy.ndarray
        The flux values in flam units. This can be a single N-length array
        or an M x N matrix.
    wavelengths: list or numpy.ndarray
        The wavelength values associated with the input flux values.
        This can be a single N-length array or an M x N matrix. If it is an
        N-length array, the same wavelength values are used for each flux_flam.
    wave_unit: astropy.units.Unit
        The unit for the wavelength values.
    flam_unit: astropy.units.Unit
        The unit for the input flux_flam values.
    fnu_unit: astropy.units.Unit
        The unit for the output flux_fnu values.

    Returns
    -------
    flux_fnu : list or np.array
        The flux values in fnu units.
    """
    flux_flam = np.asarray(flux_flam)
    wavelengths = np.asarray(wavelengths)

    # Check if we need to reshape wavelengths to match the number
    # of rows in flux_flam.
    if flux_flam.ndim > 1 and wavelengths.ndim == 1:
        wavelengths = wavelengths[None, :]

    # Check that the shapes match.
    try:
        _ = np.broadcast_shapes(flux_flam.shape, wavelengths.shape)
    except ValueError as err:
        raise ValueError(
            f"Mismatched sizes for flux_flam={flux_flam.shape} and wavelengths={wavelengths.shape}."
        ) from err

    # convert flux in flam_unit (e.g. ergs/s/cm^2/A) to fnu_unit (e.g. nJy or ergs/s/cm^2/Hz)
    input_units = (flam_unit * wave_unit * wave_unit) / const.c.unit
    flux_fnu = flux_flam * (wavelengths**2) / const.c.value
    return bulk_convert_matrix(flux_fnu, input_units, fnu_unit)


def fnu_to_flam(flux_fnu, wavelengths, *, wave_unit, flam_unit, fnu_unit):
    """
    Covert flux from f_nu unit to f_lambda unit

    Parameters
    ----------
    flux_fnu : list or numpy.ndarray
        The flux values in fnu units. This can be a single N-length array
        or an M x N matrix.
    wavelengths: list or numpy.ndarray
        The wavelength values associated with the input flux values.
        This can be a single N-length array or an M x N matrix. If it is an
        N-length array, the same wavelength values are used for each flux_fnu.
    wave_unit: astropy.units.Unit
        The unit for the wavelength values.
    flam_unit: astropy.units.Unit
        The unit for the output flux_flam values.
    fnu_unit: astropy.units.Unit
        The unit for the input flux_fnu values.

    Returns
    -------
    flux_flam : list or np.array
        The flux values in flam units.
    """
    flux_fnu = np.asarray(flux_fnu)
    wavelengths = np.asarray(wavelengths)

    # Check if we need to reshape wavelengths to match the number
    # of rows in flux_fnu.
    if flux_fnu.ndim > 1 and wavelengths.ndim == 1:
        wavelengths = wavelengths[None, :]

    # Check that the shapes match.
    try:
        _ = np.broadcast_shapes(flux_fnu.shape, wavelengths.shape)
    except ValueError as err:
        raise ValueError(
            f"Mismatched sizes for flux_fnu={flux_fnu.shape} and wavelengths={wavelengths.shape}."
        ) from err

    # convert flux in fnu_unit (e.g. nJy or ergs/s/cm^2/Hz) to flam_unit (e.g. ergs/s/cm^2/A)
    input_units = (fnu_unit * const.c.unit) / (wave_unit * wave_unit)
    flux_flam = flux_fnu * const.c.value / wavelengths**2
    return bulk_convert_matrix(flux_flam, input_units, flam_unit)



"""Wrappers for the models defined in sncosmo.

https://github.com/sncosmo/sncosmo/blob/v2.10.1/sncosmo/models.py
https://sncosmo.readthedocs.io/en/stable/models.html
"""

from astropy import units as u
from citation_compass import CiteClass

from lightcurvelynx.astro_utils.unit_utils import flam_to_fnu
from lightcurvelynx.models.physical_model import SEDModel


class SncosmoWrapperModel(SEDModel, CiteClass):
    """A wrapper for sncosmo models.

    Parameterized values include:
      * dec - The object's declination in degrees. [from BasePhysicalModel]
      * distance - The object's luminosity distance in pc. [from BasePhysicalModel]
      * ra - The object's right ascension in degrees. [from BasePhysicalModel]
      * redshift - The object's redshift. [from BasePhysicalModel]
      * t0 - The t0 of the zero phase, date. [from BasePhysicalModel]
    Additional parameterized values are used for specific sncosmo models.

    References
    ----------
    * sncosmo - https://zenodo.org/records/14714968
    * Individual models might require citation. See references in the sncosmo documentation.

    Attributes
    ----------
    source : sncosmo.Source
        The underlying source model.
    source_name : str
        The name used to set the source.
    source_param_names : list
        A list of the source model's parameters that we need to set.

    Parameters
    ----------
    source_name : str
        The name used to set the source.
    node_label : str, optional
        An identifier (or name) for the current node.
    wave_extrapolation : FluxExtrapolationModel or tuple, optional
        The extrapolation model(s) to use for wavelengths that fall outside the model's defined
        bounds. If a tuple is provided, then it is expected to be of the form (before_model, after_model)
        where before_model is the model for before the first valid wavelength and after_model is
        the model for after the last valid wavelength. If None is provided the model will not try to
        extrapolate, but rather call compute_sed() for all wavelengths.
    time_extrapolation : FluxExtrapolationModel or tuple, optional
        The extrapolation model(s) to use for times that fall outside the model's defined
        bounds. If a tuple is provided, then it is expected to be of the form (before_model, after_model)
        where before_model is the model for before the first valid time and after_model is
        the model for after the last valid time. If None is provided the model will not try to
        extrapolate, but rather call compute_sed() for all times.
    seed : int, optional
        The seed for a random number generator.
    **kwargs : dict, optional
        Any additional keyword arguments.
    """

    # A class variable for the units so we are not computing them each time.
    _FLAM_UNIT = u.erg / u.second / u.cm**2 / u.AA

    def __init__(
        self,
        source_name,
        node_label=None,
        wave_extrapolation=None,
        time_extrapolation=None,
        seed=None,
        **kwargs,
    ):
        try:
            from sncosmo.models import get_source
        except ImportError as err:  # pragma: no cover
            raise ImportError(
                "sncosmo package is not installed be default. To use the SncosmoWrapperModel, "
                "please install sncosmo. For example, you can install it with "
                "`pip install sncosmo` or `conda install conda-forge::sncosmo`."
            ) from err

        # We explicitly ask for and pass along the PhysicalModel parameters such
        # as node_label and wave_extrapolation so they do not go into kwargs
        # and get added to the sncosmo model below.
        super().__init__(
            node_label=node_label,
            wave_extrapolation=wave_extrapolation,
            time_extrapolation=time_extrapolation,
            seed=seed,
            **kwargs,
        )
        self.source_name = source_name
        self.source = get_source(source_name)

        # Use the kwargs to initialize the sncosmo model's parameters.
        self.source_param_names = []
        for key, value in kwargs.items():
            if key not in self.setters:
                self.add_parameter(key, value, description="Parameter for sncosmo model.")
            if key in self.source.param_names:
                self.source_param_names.append(key)

    @property
    def param_names(self):
        """Return a list of the model's parameter names."""
        return self.source.param_names

    @property
    def parameter_values(self):
        """Return a list of the model's parameter values."""
        return self.source.parameters

    def minphase(self, **kwargs):
        """Get the minimum phase of the model (in days relative to t0).

        Parameters
        ----------
        **kwargs : dict
            Additional keyword arguments, not used in this method.

        Returns
        -------
        minphase : float or None
            The minimum phase of the model (in days relative to t0) or None
            if the model does not have a defined minimum phase.
        """
        return self.source.minphase()

    def maxphase(self, **kwargs):
        """Get the maximum phase of the model (in days relative to t0).

        Parameters
        ----------
        **kwargs : dict
            Additional keyword arguments, not used in this method.

        Returns
        -------
        maxphase : float or None
            The maximum phase of the model (in days relative to t0) or None
            if the model does not have a defined maximum phase.
        """
        return self.source.maxphase()

    def minwave(self, **kwargs):
        """Get the minimum wavelength of the model.

        Parameters
        ----------
        **kwargs : dict
            Additional keyword arguments, not used in this method.

        Returns
        -------
        minwave : float or None
            The minimum wavelength of the model (in angstroms) or None
            if the model does not have a defined minimum wavelength.
        """
        return self.source.minwave()

    def maxwave(self, **kwargs):
        """Get the maximum wavelength of the model.

        Parameters
        ----------
        **kwargs : dict
            Additional keyword arguments, not used in this method.

        Returns
        -------
        maxwave : float or None
            The maximum wavelength of the model (in angstroms) or None
            if the model does not have a defined maximum wavelength.
        """
        return self.source.maxwave()

    def _update_sncosmo_model_parameters(self, graph_state):
        """Update the parameters for the wrapped sncosmo model."""
        local_params = graph_state.get_node_state(self.node_string, 0)
        sn_params = {}
        for name in self.source_param_names:
            sn_params[name] = local_params[name]
        self.source.set(**sn_params)

    def get(self, name):
        """Get the value of a specific parameter.

        Parameters
        ----------
        name : str
            The name of the parameter.

        Returns
        -------
        The parameter value.
        """
        return self.source.get(name)

    def set(self, **kwargs):
        """Set the parameters of the model.

        These must all be constants to be compatible with sncosmo.

        Parameters
        ----------
        **kwargs : dict
            The parameters to set and their values.
        """
        for key, value in kwargs.items():
            if hasattr(self, key):
                self.set_parameter(key, value)
            else:
                self.add_parameter(key, value, description="Parameter for sncosmo model.")
            if key not in self.source_param_names:
                self.source_param_names.append(key)
        self.source.set(**kwargs)

    def _sample_helper(self, graph_state, seen_nodes, rng_info=None):
        """Internal recursive function to sample the model's underlying parameters
        if they are provided by a function or ParameterizedNode.

        Calls ParameterNode's _sample_helper() then updates the parameters
        for the sncosmo model.

        Parameters
        ----------
        graph_state : GraphState
            An object mapping graph parameters to their values. This object is modified
            in place as it is sampled.
        seen_nodes : dict
            A dictionary mapping nodes seen during this sampling run to their ID.
            Used to avoid sampling nodes multiple times and to validity check the graph.
        num_samples : int
            A count of the number of samples to compute.
            Default: 1
        rng_info : numpy.random._generator.Generator, optional
            A given numpy random number generator to use for this computation. If not
            provided, the function uses the node's random number generator.

        Raises
        ------
        Raise a ValueError the sampling encounters a problem with the order of dependencies.
        """
        super()._sample_helper(graph_state, seen_nodes, rng_info=rng_info)
        self._update_sncosmo_model_parameters(graph_state)

    def mask_by_time(self, times, graph_state=None):
        """Compute a mask for whether a given time is of interest for a given object.
        For example, a user can use this function to generate a mask to include
        only the observations of interest for a window around the supernova.

        Parameters
        ----------
        times : numpy.ndarray
            A length T array of observer frame timestamps in MJD.
        graph_state : GraphState, optional
            An object mapping graph parameters to their values.

        Returns
        -------
        time_mask : numpy.ndarray
            A length T array of Booleans indicating whether the time is of interest.
        """
        if graph_state is None:
            raise ValueError("graph_state needed to compute mask_by_time")

        z = self.get_param(graph_state, "redshift", 0.0)
        if z is None:
            z = 0.0

        t0 = self.get_param(graph_state, "t0", 0.0)
        if t0 is None:
            t0 = 0.0

        # Compute the mask.
        good_times = (times > t0 + self.source.minphase() * (1.0 + z)) & (
            times < t0 + self.source.maxphase() * (1.0 + z)
        )
        return good_times

    def compute_sed(self, times, wavelengths, graph_state=None, **kwargs):
        """Draw effect-free observations for this object.

        Parameters
        ----------
        times : numpy.ndarray
            A length T array of rest frame timestamps.
        wavelengths : numpy.ndarray, optional
            A length N array of wavelengths (in angstroms).
        graph_state : GraphState
            An object mapping graph parameters to their values.
        **kwargs : dict, optional
           Any additional keyword arguments.

        Returns
        -------
        flux_density : numpy.ndarray
            A length T x N matrix of SED values (in nJy).
        """
        params = self.get_local_params(graph_state)
        self._update_sncosmo_model_parameters(graph_state)

        # Query the model and convert the output to nJy.
        phase = times - params["t0"]
        model_flam = self.source.flux(phase, wavelengths)
        model_fnu = flam_to_fnu(
            model_flam,
            wavelengths,
            wave_unit=u.AA,
            flam_unit=self._FLAM_UNIT,
            fnu_unit=u.nJy,
        )
        return model_fnu



"""A wrapper class for representing detector footprints (which are stored as
astropy PixelRegion). This class provides methods for checking if points are within
the footprint and for plotting the footprint."""

import matplotlib.pyplot as plt
import numpy as np
from astropy import units as u
from astropy.coordinates import SkyCoord
from astropy.wcs import WCS
from regions import PixCoord, RectanglePixelRegion, RectangleSkyRegion, SkyRegion


class DetectorFootprint:
    """A wrapper class for representing detector footprints.

    Attributes
    ----------
    region : astropy.regions.PixelRegion
        The astropy PixelRegion representing the footprint.
    wcs : astropy.wcs.WCS or None
        The WCS associated with the region, if any.

    Parameters
    ----------
    region : astropy.regions.SkyRegion or astropy.regions.PixelRegion
        The astropy SkyRegion or PixelRegion representing the footprint. SkyRegions
        will be converted to PixelRegions using the provided WCS or a default WCS.
    wcs : astropy.wcs.WCS or None
        The WCS associated with the region, if any.
    pixel_scale : float or None
        The pixel scale in arcseconds/pixel, this is required if no WCS is provided.
    center_pixels : tuple of float, optional
        The pixel coordinates of the center of the detector. Default is (0.5, 0.5) for
        the center of the (0, 0) pixel. This is only used if no WCS is provided and
        a default WCS is created.
    """

    def __init__(self, region, *, wcs=None, pixel_scale=None, center_pixels=(0.5, 0.5), **kwargs):
        # Create a default WCS if none is provided.
        if wcs is None:
            if pixel_scale is None:
                raise ValueError("Either wcs or pixel_scale must be provided.")
            if pixel_scale <= 0:
                raise ValueError("pixel_scale must be positive.")
            pixel_scale_deg = pixel_scale / 3600.0  # Convert to degrees/pixel

            # Create a simple TAN WCS centered on (0.0, 0.0) with the given pixel scale.
            wcs = WCS(naxis=2)
            wcs.wcs.ctype = ["RA---TAN", "DEC--TAN"]
            wcs.wcs.crval = [0.0, 0.0]  # Centered on RA=0.0, dec=0.0
            wcs.wcs.crpix = [center_pixels[0], center_pixels[1]]
            wcs.wcs.cdelt = [-pixel_scale_deg, pixel_scale_deg]  # The given pixel scale in degrees/pixel
        self.wcs = wcs

        # Store the region as a pixel region, since we will always need to do a conversion
        # as part of the contains method otherwise.
        self.region = region
        if isinstance(self.region, SkyRegion):
            self.region = self.region.to_pixel(self.wcs)

        # The bounding box must contain the origin (0,0) since the footprint is defined
        # relative to the center of the detector.
        bbox = self.region.bounding_box
        if not (bbox.ixmin <= 0 <= bbox.ixmax and bbox.iymin <= 0 <= bbox.iymax):
            raise ValueError("The bounding box of the region must contain the origin (0,0).")

    @classmethod
    def from_sky_rect(cls, width, height, wcs=None, pixel_scale=None, **kwargs):
        """Create a rectangular footprint in degrees.

        Parameters
        ----------
        width : float
            Width of the rectangle in degrees.
        height : float
            Height of the rectangle in degrees.
        wcs : astropy.wcs.WCS, optional
            The WCS associated with the region. If None, a default WCS will be created.
        pixel_scale : float, optional
            The pixel scale in arcseconds/pixel, this is required if no WCS is provided.
        **kwargs : dict
            Additional keyword arguments to pass to the RectangleSkyRegion constructor.

        Returns
        -------
        DetectorFootprint
            The rectangular detector footprint.
        """
        center = SkyCoord(ra=0.0, dec=0.0, unit="deg", frame="icrs")
        region = RectangleSkyRegion(
            center=center,
            width=width * u.deg,
            height=height * u.deg,
            angle=0.0 * u.deg,
            **kwargs,
        )
        return cls(region, wcs=wcs, pixel_scale=pixel_scale)

    @classmethod
    def from_pixel_rect(cls, width, height, wcs=None, pixel_scale=None, **kwargs):
        """Create a rectangular footprint in pixels.

        Parameters
        ----------
        width : float
            Width of the rectangle in pixels.
        height : float
            Height of the rectangle in pixels.
        wcs : astropy.wcs.WCS, optional
            The WCS associated with the region. If None, a default WCS will be created.
        pixel_scale : float, optional
            The pixel scale in arcseconds/pixel, this is required if no WCS is provided.
        **kwargs : dict
            Additional keyword arguments to pass to the RectangleSkyRegion constructor.

        Returns
        -------
        DetectorFootprint
            The rectangular detector footprint.
        """
        center = PixCoord(x=0.0, y=0.0)
        region = RectanglePixelRegion(
            center=center,
            width=width,
            height=height,
            angle=0.0 * u.deg,
            **kwargs,
        )
        return cls(region, wcs=wcs, pixel_scale=pixel_scale)

    @staticmethod
    def rotate_to_center(ra, dec, center_ra, center_dec, *, rotation=None):
        """Transform the given points, represented by (ra, dec) in degrees,
        to a local coordinate system centered on (center_ra, center_dec), accounting
        for rotation if specified.

        Note
        ----
        This method is vectorized so it can handle an array of points and an (equally
        sized) array of center points. The transformation is done on each point for the
        corresponding center point.

        Parameters
        ----------
        ra : np.ndarray
            Right ascension in degrees.
        dec : np.ndarray
            Declination in degrees.
        center_ra : np.ndarray
            Center right ascension of the detector in degrees.
        center_dec : np.ndarray
            Center declination of the detector in degrees.
        rotation : np.ndarray, optional
            The rotation angle of the detector for each pointing in degrees clockwise.
            Used to represent non-axis-aligned footprints.

        Returns
        -------
        tuple of np.ndarray
            Transformed (lon, lat) offsets from the center in degrees.
        """
        if rotation is None:
            rotation = np.zeros_like(ra)

        target = SkyCoord(ra=ra * u.deg, dec=dec * u.deg, frame="icrs")
        origin = SkyCoord(ra=center_ra * u.deg, dec=center_dec * u.deg, frame="icrs")
        offset_frame = origin.skyoffset_frame(rotation=rotation * u.deg)
        offset = target.transform_to(offset_frame)

        # Center everything on 0.0
        lon_t = offset.lon.deg
        lon_t[lon_t > 180.0] -= 360.0

        lat_t = offset.lat.deg
        lat_t[lat_t > 90.0] -= 180.0

        return lon_t, lat_t

    def sky_to_pixel(self, ra, dec, center_ra, center_dec, *, rotation=None):
        """Transform sky coordinates (ra, dec) to pixel coordinates.

        Parameters
        ----------
        ra : float or array-like
            Right ascension in degrees.
        dec : float or array-like
            Declination in degrees.
        center_ra : float
            Center right ascension of the detector in degrees.
        center_dec : float
            Center declination of the detector in degrees.
        rotation : float or array-like, optional
            The rotation angle of the detector for each pointing in degrees clockwise.
            Used to represent non-axis-aligned footprints.

        Returns
        -------
        tuple of np.ndarray
            Pixel coordinates (x, y) corresponding to the input sky coordinates.
        """
        lon_t, lat_t = self.rotate_to_center(ra, dec, center_ra, center_dec, rotation=rotation)
        sky_pts = SkyCoord(lon_t, lat_t, unit="deg")
        pixel_x, pixel_y = self.wcs.world_to_pixel(sky_pts)
        return pixel_x, pixel_y

    def contains(self, ra, dec, center_ra, center_dec, *, rotation=None):
        """Check that given points, represented by (ra, dec) in degrees,
        are within the footprint.

        Parameters
        ----------
        ra : float or array-like
            Right ascension in degrees.
        dec : float or array-like
            Declination in degrees.
        center_ra : float or array-like
            Center right ascension of the detector in degrees.
        center_dec : float or array-like
            Center declination of the detector in degrees.
        rotation : np.ndarray, optional
            The rotation angle of the detector for each pointing in degrees clockwise.
            Used to represent non-axis-aligned footprints.

        Returns
        -------
        bool or array-like of bool
            True if the point is within the footprint, False otherwise.
        """
        scalar_data = np.isscalar(ra)

        # Make all inputs into arrays and confirm they have the same shape.
        ra = np.atleast_1d(ra)
        dec = np.atleast_1d(dec)
        if ra.shape != dec.shape:
            raise ValueError("ra and dec must have the same shape.")

        center_ra = np.atleast_1d(center_ra)
        center_dec = np.atleast_1d(center_dec)
        if center_ra.shape != ra.shape:
            if center_ra.shape != (1,):
                raise ValueError("center_ra must have the same shape as ra and dec.")
            else:
                center_ra = np.full(ra.shape, center_ra[0])
        if center_dec.shape != dec.shape:
            if center_dec.shape != (1,):
                raise ValueError("center_dec must have the same shape as ra and dec.")
            else:
                center_dec = np.full(dec.shape, center_dec[0])

        # Rotation is optional, but if provided, it must match the shape of ra and dec.
        if rotation is not None:
            rotation = np.atleast_1d(rotation)
            if rotation.shape != ra.shape:
                if rotation.shape != (1,):
                    raise ValueError("rotation must have the same shape as ra and dec.")
                else:
                    rotation = np.full(ra.shape, rotation[0])

        # Rotate the points to be relative to the center and convert to pixel coordinates.
        pixel_x, pixel_y = self.sky_to_pixel(ra, dec, center_ra, center_dec, rotation=rotation)
        pix_coord = PixCoord(x=pixel_x, y=pixel_y)

        # Check if each point is within the footprint.
        result = self.region.contains(pix_coord)
        if scalar_data:
            return result[0]
        return result

    def compute_radius(self):
        """Compute an approximate bounding radius of the footprint in degrees.

        Note
        ----
        This radius is based on the bounding box of the footprint (with a small
        pixel padding), so it will be larger than the actual radius of the footprint.

        Returns
        -------
        float
            The radius of the footprint in degrees.
        """
        # Compute the corners of the bounding box (using a pixel buffer to handle different
        # centering schemes) and convert to sky coordinates.
        bbox = self.region.bounding_box
        corner_pix_x = np.array([bbox.ixmin - 1.0, bbox.ixmin - 1.0, bbox.ixmax + 1.0, bbox.ixmax + 1.0])
        corner_pix_y = np.array([bbox.iymin - 1.0, bbox.iymax + 1.0, bbox.iymax + 1.0, bbox.iymin - 1.0])
        corner_ra, corner_dec = self.wcs.pixel_to_world_values(corner_pix_x, corner_pix_y)
        sky_corners = SkyCoord(ra=corner_ra, dec=corner_dec, unit="deg", frame="icrs")

        center = SkyCoord(ra=0.0, dec=0.0, unit="deg", frame="icrs")
        separations = center.separation(sky_corners)
        return np.max(separations.deg)

    def compute_sky_region(self, center_ra, center_dec, *, rotation=0.0):
        """Compute the sky region of the footprint at the given center position.

        Parameters
        ----------
        center_ra : float
            Center right ascension of the detector in degrees.
        center_dec : float
            Center declination of the detector in degrees.
        rotation : float, optional
            The rotation angle of the detector for each pointing in degrees clockwise.
            Used to represent non-axis-aligned footprints. Default is 0.0.

        Returns
        -------
        astropy.regions.SkyRegion
            The sky region representing the footprint at the given center position.
        astropy.wcs.WCS
            The WCS representing the footprint at the given center position.
        """
        # Perform the rotation in pixel coordinates first.
        if rotation != 0.0:
            center = PixCoord(x=0.0, y=0.0)
            pixel_region = self.region.rotate(center, rotation * u.deg)
        else:
            pixel_region = self.region

        # Convert to sky region and set the center.
        sky_region = pixel_region.to_sky(self.wcs)
        sky_region.center = SkyCoord(ra=center_ra, dec=center_dec, unit="deg", frame="icrs")

        # Create a copy of the WCS with the new center.
        sky_wcs = self.wcs.deepcopy()
        sky_wcs.wcs.crval = [center_ra, center_dec]

        return sky_region, sky_wcs

    def plot(
        self,
        *,
        ax=None,
        figure=None,
        center_ra=0,
        center_dec=0,
        rotation=0,
        point_ra=None,
        point_dec=None,
        **kwargs,
    ):
        """Plot the footprint using matplotlib and an optional set of points to overlay.

        Parameters
        ----------
        ax : matplotlib.pyplot.Axes or None, optional
            Axes, If None, a new axes will be created. None by default.
        figure : matplotlib.pyplot.Figure or None
            Figure, If None, a new figure will be created. None by default.
        center_ra : float, optional
            Center right ascension of the detector in degrees. Default is 0.
        center_dec : float, optional
            Center declination of the detector in degrees. Default is 0.
        rotation : np.ndarray, optional
            The rotation angle of the detector for each pointing in degrees clockwise.
            Used to represent non-axis-aligned footprints.
            Default is 0.
        point_ra : array-like or None, optional
            Right ascension of points to overlay in degrees. None by default.
        point_dec : array-like or None, optional
            Declination of points to overlay in degrees. None by default.
        **kwargs : dict
            Optional parameters to pass to the plotting function
        """
        if ax is None:
            if figure is None:
                figure = plt.figure()
            ax = figure.add_axes([0, 0, 1, 1])

        # Plot the bounds of the footprint.
        artist = self.region.as_artist()
        ax.add_artist(artist)

        # Get the bounding box in pixel coordinates. Expand out to ensure the full
        # footprint is visible.
        bbox = self.region.bounding_box
        width = bbox.ixmax - bbox.ixmin
        height = bbox.iymax - bbox.iymin
        xmin = bbox.ixmin - 0.5 * width
        xmax = bbox.ixmax + 0.5 * width
        ymin = bbox.iymin - 0.5 * height
        ymax = bbox.iymax + 0.5 * height

        # If points are provided, overlay them on the plot.
        if point_ra is not None and point_dec is not None:
            # Compute the transformed pixel coordinates relative to the center.
            pixel_x, pixel_y = self.sky_to_pixel(
                point_ra, point_dec, center_ra, center_dec, rotation=rotation
            )

            # Since we have already transformed the points to be relative to the center,
            # we can check if they are within the footprint without further transformation.
            isin = self.region.contains(PixCoord(x=pixel_x, y=pixel_y))

            ax.scatter(pixel_x[~isin], pixel_y[~isin], color="red", marker="x", label="Outside Footprint")
            ax.scatter(pixel_x[isin], pixel_y[isin], color="green", marker="o", label="Inside Footprint")
            ax.legend()

            # Adjust the bounding box to include all the points.
            if xmin < pixel_x.min():
                xmin = pixel_x.min() - 2.0
            if xmax > pixel_x.max():
                xmax = pixel_x.max() + 2.0
            if ymin < pixel_y.min():
                ymin = pixel_y.min() - 2.0
            if ymax > pixel_y.max():
                ymax = pixel_y.max() + 2.0

        ax.set_title("Detector Footprint")
        ax.set_xlabel("Pixel X")
        ax.set_ylabel("Pixel Y")
        ax.set_aspect("equal")
        ax.set_xlim([xmin, xmax])
        ax.set_ylim([ymin, ymax])
        plt.show()



"""Model that generate the SED or bandflux of a source based on given observer frame
light curves of fluxes in each band.

If we are generating the bandfluxes directly, the models interpolate the given light curves
at the requested times and filters. If we are generating an SED for a given set of
wavelengths, the model computes a box-shaped SED basis function for each filter that
will produce the same bandflux after being passed through the passband filter.

Note: If you are interested in generating SED-level data, use the SEDTemplateModel in
src/lightcurvelynx/models/sed_template_model.py instead.
"""

import logging
from abc import ABC

import matplotlib.pyplot as plt
import numpy as np
from citation_compass import cite_inline
from tqdm import tqdm

from lightcurvelynx.astro_utils.mag_flux import mag2flux
from lightcurvelynx.astro_utils.passbands import Passband, PassbandGroup
from lightcurvelynx.astro_utils.sed_basis_models import SEDBasisModel
from lightcurvelynx.consts import lsst_filter_plot_colors
from lightcurvelynx.math_nodes.given_sampler import GivenValueSampler, GivenValueSelector
from lightcurvelynx.models.physical_model import BandfluxModel
from lightcurvelynx.utils.io_utils import read_lclib_data

logger = logging.getLogger(__name__)


class LightcurveBandData:
    """A class to hold data for a single model light curve defined at the band level (a set of
    fluxes over time for each filter).

    Data can be passed in as fluxes (in nJy) or AB magnitudes (if magnitudes_in=True), but
    is always stored internally as fluxes.

    Attributes
    ----------
    lightcurves : dict
        A dictionary mapping filter names to a 2D array of the bandfluxes in that filter,
        where the first column is time (in days from the reference time of the light curve),
        the second column is the bandflux (in nJy), and an optional third column is fluxerror.
    lc_data_t0 : float
        The reference epoch of the input light curve. This is the time stamp of the input
        array that will correspond to t0 in the model. For periodic light curves, this either
        must be set to the first time of the light curve or set as 0.0 to automatically
        derive the lc_data_t0 from the light curve.
    period : float or None
        The period of the light curve in days. If the light curve is not periodic,
        then this value is set to None.
    min_times : dict
        A dictionary mapping filter names to the minimum time of the light curve
        in that filter (shifted to be relative to the reference epoch of the light curve).
    max_times : dict
        A dictionary mapping filter names to the maximum time of the light curve
        in that filter (shifted to be relative to the reference epoch of the light curve).
    baseline : dict
        A dictionary of baseline bandfluxes for each filter (in nJy). This is only used
        for non-periodic light curves when they are not active.
    has_time_bounds : bool
        Whether the model has time bounds where it is valid. This is True for non-periodic
        light curves without a baseline, and False otherwise.

    Parameters
    ----------
    lightcurves : dict or numpy.ndarray
        The light curves can be passed as either:
        1) a dictionary mapping filter names to a (T, 2) array of the bandfluxes in that filter
        where the first column is time and the second column is the light curve values, or
        2) a numpy array of shape (T, 3) array where the first column is time (in days), the
        second column is the light curve values, and the third column is the filter.
        The light curve values can be either fluxes (in nJy) or AB magnitudes (if magnitudes_in=True).
    lc_data_t0 : float
        The reference epoch of the input light curve. The model will be shifted
        to the model's lc_data_t0 when computing fluxes.  For periodic light curves, this either
        must be set to the first time of the light curve or as 0.0 to automatically
        derive the lc_data_t0 from the light curve.
    periodic : bool
        Whether the light curve is periodic. If True, the model will assume that
        the light curve repeats every period.
        Default: False
    magnitudes_in : bool
        Whether the input light curves are in AB magnitudes (True) or fluxes (False).
    baseline : dict or None
        A dictionary of baseline bandfluxes or AB magnitudes for each filter. This is only used
        for non-periodic light curves when they are not active.
        Default: None
    """

    def __init__(
        self,
        lightcurves,
        lc_data_t0,
        *,
        periodic=False,
        magnitudes_in=False,
        baseline=None,
    ):
        if lc_data_t0 is None:
            raise ValueError("lc_data_t0 must be provided and cannot be None.")
        self.lc_data_t0 = lc_data_t0
        self.period = None

        if isinstance(lightcurves, dict):
            # Make a copy of the light curves to avoid modifying the original data.
            self.lightcurves = {filter: lc.copy() for filter, lc in lightcurves.items()}
        elif isinstance(lightcurves, np.ndarray):
            if lightcurves.shape[1] != 3:
                raise ValueError("Light curves array must have 3 columns: time, flux, and filter.")

            # Break up the light curves by filter.
            self.lightcurves = {}
            filters = np.unique(lightcurves[:, 2])
            for filter in filters:
                filter_mask = lightcurves[:, 2] == filter
                filter_times = lightcurves[filter_mask, 0].astype(float)
                filter_bandflux = lightcurves[filter_mask, 1].astype(float)
                self.lightcurves[str(filter)] = np.column_stack((filter_times, filter_bandflux))
        else:
            raise TypeError(
                "Unknown type for light curve input. Must be dict, numpy array, or astropy Table."
            )

        # Do basic validation of the light curves and shift them so that the time
        # at lc_data_t0 is mapped to 0.0. Convert from AB magnitudes to fluxes if needed.
        for filter, lc in self.lightcurves.items():
            if len(lc.shape) != 2 or (lc.shape[1] != 2 and lc.shape[1] != 3):
                raise ValueError(f"Lightcurve {filter} must have either 2 or 3 columns.")
            if lc.shape[1] == 3:
                lc = lc[:, :2]  # Drop the error column if present.
            if not np.all(np.diff(lc[:, 0]) > 0):
                raise ValueError(f"Lightcurve {filter}'s times are not in sorted order.")

            # Shift the light curve times to be relative to lc_data_t0.
            lc[:, 0] -= self.lc_data_t0

            # Convert from magnitudes to fluxes if needed.
            if magnitudes_in:
                lc[:, 1] = mag2flux(lc[:, 1])

        # Store the minimum and maximum times for each light curve. This is done after
        # validating periodicity in case we needed to adjust the light curve start times.
        if periodic:
            self._validate_periodicity()
        self.min_times = {filter: lc[0, 0] for filter, lc in self.lightcurves.items()}
        self.max_times = {filter: lc[-1, 0] for filter, lc in self.lightcurves.items()}

        # If the model is periodic or has a given baseline, it is considered valid
        # outside the minimum and maximum times of each light curve.
        self.has_time_bounds = (not periodic) and baseline is None

        # Store the baseline values for each filter. If the baseline is provided,
        # make sure it contains all of the filters. If no baseline is provided,
        # set the baseline to 0.0 for each filter.
        if baseline is None:
            self.baseline = {filter: 0.0 for filter in self.lightcurves}
        else:
            for filter in self.lightcurves:
                if filter not in baseline:
                    raise ValueError(f"Baseline value for filter {filter} is missing.")
            self.baseline = baseline

        # Convert the baseline from magnitudes to fluxes if needed.
        if magnitudes_in:
            for filter in self.baseline:
                self.baseline[filter] = mag2flux(self.baseline[filter])

    def __len__(self):
        """Get the number of light curves."""
        return len(self.lightcurves)

    @property
    def filters(self):
        """Get the list of filters in the lightcurves."""
        return list(self.lightcurves.keys())

    def _validate_periodicity(self):
        """Check that the light curves meet the requirements for periodic models:
        - All light curves must be sampled at the same times.
        - The light curves must have a non-zero period.
        - The value at the start and end of each light curve must be the same.
        """
        all_lcs = list(self.lightcurves.values())
        if len(all_lcs) == 0:
            raise ValueError("Periodic light curve models must have at least one light curve.")
        if len(all_lcs[0]) < 2:
            raise ValueError("All periodic light curves must have at least two time points.")

        # Check that all light curves are sampled at the same times and the first value
        # matches the last value.
        num_curves = len(all_lcs)
        for i in range(num_curves):
            if not np.allclose(all_lcs[i][:, 0], all_lcs[0][:, 0]):
                raise ValueError("All light curves in a periodic model must be sampled at the same times.")
            if not np.allclose(all_lcs[i][0, 1], all_lcs[i][-1, 1]):
                raise ValueError("All periodic light curves must have the same value at the start and end.")

        # Check that all light curves have a non-zero period.
        self.period = all_lcs[0][-1, 0] - all_lcs[0][0, 0]
        if self.period <= 0.0:
            raise ValueError("The period of the light curve must be positive.")

        # Shift all the lightcurves so they start at 0 (to make the math easier)
        # and record the offset as lc_data_t0.
        if not np.isclose(all_lcs[0][0, 0], 0.0):
            if self.lc_data_t0 != 0.0:
                raise ValueError(
                    "For periodic models, lc_data_t0 must either be set to the first time "
                    f"or automatically derived. Found lc_data_t0={self.lc_data_t0}."
                )

            self.lc_data_t0 = all_lcs[0][0, 0]
            for lc in self.lightcurves.values():
                lc[:, 0] -= self.lc_data_t0

    @classmethod
    def from_lclib_table(cls, lightcurves_table, *, forced_lc_t0=None, filters=None):
        """Break up a light curves table in LCLIB format into a LightcurveBandData instance.
        This function expects the table to have a "time" column, an optional "type" column,
        and a column for each filter. The "type" column should use "S" for source observation
        and "T" for template (background) observation.

        Parameters
        ----------
        lightcurves_table : astropy.table.Table
            A table with a "time" column, optional "type" column, and a column for each filter.
            If the type column is present it should use "S" for source observation and "T"
            for template (background) observation.
        forced_lc_t0 : float
            By default we use the LCLIB convention of storing the light curves so the first time
            corresponds to the reference epoch (lc_data_t0) of the light curve. This can be overridden
            by providing a value for forced_lc_t0.
            Default: None
        filters : list of str or None
            A list of filters to use for the light curves. If None, all filters will be used.
            Used to select a subset of filters.
            Default: None
        """
        if "time" not in lightcurves_table.colnames:
            raise ValueError("Light curves table must have a 'time' column.")

        # Extract the name of the filters from the table column names.
        filter_cols = [col for col in lightcurves_table.colnames if col != "time" and col != "type"]
        if filters is None:
            filters = filter_cols
        else:
            to_keep = set(filter_cols) & set(filters)
            filters = list(to_keep)
        if len(filters) == 0:
            raise ValueError("Light curves table must have at least one filter column.")

        # Check if there are baseline curves to extract and filter them out of the
        # light curves table. Use a default to 0.0 for each filter if no baselines are found.
        baseline = {filter: 0.0 for filter in filters}
        if "type" in lightcurves_table.colnames:
            obs_mask = lightcurves_table["type"] == "S"
            if np.any(~obs_mask):
                tmp_table = lightcurves_table[~obs_mask]
                if len(tmp_table) > 1:
                    logger.warning(
                        "Multiple template (background) observations found in light curves table. "
                        "The light curve will only use the first one for baseline values."
                    )
                baseline = {filter: mag2flux(tmp_table[filter][0]) for filter in filters}
            lightcurves_table = lightcurves_table[obs_mask]

        # Determine the reference epoch of the light curve (lc_data_t0).
        lc_data_t0 = np.min(lightcurves_table["time"]) if forced_lc_t0 is None else forced_lc_t0

        # Convert the Table to a dictionary of lightcurves.
        lightcurves = {}
        for filter in filters:
            filter_times = lightcurves_table["time"].astype(float)
            filter_bandflux = mag2flux(lightcurves_table[filter].astype(float))
            lightcurves[str(filter)] = np.column_stack((filter_times, filter_bandflux))

        # Check the metadata for periodicity information.
        recur_class = lightcurves_table.meta.get("RECUR_CLASS", "")
        if recur_class == "PERIODIC" or recur_class == "RECUR-PERIODIC":
            periodic = True
            baseline = None  # Baseline is not used for periodic light curves.
        elif recur_class == "RECUR-NONPERIODIC":
            periodic = False
            logger.warning(
                "Recurring non-periodic light curves are treated as non-recurring within LightCurveLynx."
            )
        elif recur_class == "NON-RECUR":
            periodic = False
        elif recur_class == "":
            periodic = False
            logger.warning(
                "No RECUR_CLASS metadata found in light curves table. Using non-periodic light curves."
            )
        else:
            raise ValueError(
                f"Unknown RECUR_CLASS value in light curves table metadata: {recur_class}. "
                "Expected 'PERIODIC', 'RECUR-PERIODIC', 'RECUR-NONPERIODIC', or 'NON-RECUR'."
            )

        # If the light curves are periodic, make sure they start and end at the same value.
        if periodic:
            all_match = True
            for lc in lightcurves.values():
                all_match &= np.isclose(lc[0, 1], lc[-1, 1])

            # Insert a value to wrap. This should be a bit after the last time
            # and have the same value as the first time.
            if not all_match:
                dt = lightcurves_table["time"][-1] - lightcurves_table["time"][0]
                ave_dt = dt / (len(lightcurves_table) - 1)
                new_end = lightcurves_table["time"][-1] + ave_dt

                for filter, lc in lightcurves.items():
                    lc = np.vstack((lc, [new_end, lc[0, 1]]))
                    lightcurves[filter] = lc

        return cls(lightcurves, lc_data_t0, periodic=periodic, baseline=baseline)

    def evaluate_bandfluxes(self, times, filter):
        """Get the bandflux values for a given filter at the specified times. These can
        be multiplied by a basis SED function to produce estimated SED values
        for the given filter at the specified times or can be used directly as bandfluxes.

        Parameters
        ----------
        times : numpy.ndarray
            A length T array of times (in days) at which to compute the SED values. These should
            be shift to be relative to the light curve's lc_data_t0.
        filter : str
            The name of the filter for which to compute the SED values.

        Returns
        -------
        values : numpy.ndarray
            A length T array of bandpass fluxes for the specified filter at the given times.
        """
        if filter not in self.lightcurves:
            raise ValueError(f"Filter {filter} not found in light curves.")
        lightcurve = self.lightcurves[filter]

        # If the light curve is periodic, wrap the times around the period.
        if self.period is not None:
            times = times % self.period

        # Start with an array of all baseline values.
        values = np.full(len(times), self.baseline.get(filter, 0.0))

        # For the times that overlap with the light curve, interpolate the light curve values.
        overlap = (times >= self.min_times[filter]) & (times <= self.max_times[filter])
        values[overlap] = np.interp(
            times[overlap],  # The query times
            lightcurve[:, 0],  # The light curve times for this passband filter
            lightcurve[:, 1],  # The light curve flux densities for this passband filter
            left=0.0,  # Do not extrapolate in time
            right=0.0,  # Do not extrapolate in time
        )

        return values

    def plot_lightcurves(self, times=None, ax=None, figure=None):
        """Plot the underlying light curves. This is a debugging
        function to help the user understand the SEDs produced by this
        model.

        Parameters
        ----------
        times : numpy.ndarray or None, optional
            An array of timestamps at which to plot the light curves.
            If None, the function uses the timestamps from each light curve.
        ax : matplotlib.pyplot.Axes or None, optional
            Axes, None by default.
        figure : matplotlib.pyplot.Figure or None
            Figure, None by default.
        """
        if ax is None:
            if figure is None:
                figure = plt.figure()
            ax = figure.add_axes([0, 0, 1, 1])

        # Plot each passband.
        for filter_name, filter_curve in self.lightcurves.items():
            # Check if we need to use the query times.
            if times is None:
                plot_times = filter_curve[:, 0]
                plot_values = filter_curve[:, 1]
            else:
                plot_times = times
                plot_values = np.interp(times, filter_curve[:, 0], filter_curve[:, 1], left=0.0, right=0.0)

            color = lsst_filter_plot_colors.get(filter_name, "black")
            ax.plot(plot_times, plot_values, color=color, label=filter_name)

        # Set the x and y axis labels.
        ax.set_xlabel("Time (days)")
        ax.set_ylabel("Filter value (nJy)")
        ax.set_title("Underlying Light Curves")
        ax.legend()


class BaseLightcurveBandTemplateModel(BandfluxModel, ABC):
    """A base class for light curve template models. This class is not meant to be used directly,
    but rather as a base for other light curve template models that may have additional functionality.
    It provides the basic structure (primarily SED basis functions) and validation for
    light curve-based SED models.

    The set of passbands used to configure the model MUST be the same as used
    to generate the SED (the wavelengths must match).

    Parameterized values include:
      * dec - The object's declination in degrees.
      * ra - The object's right ascension in degrees.
      * t0 - The t0 of the zero phase (if applicable), date.

    Attributes
    ----------
    sed_basis: SEDBasisModel, optional
        An SEDBasisModel mapping representing the fake SED basis functions for each filter.
        Only generated if passbands are provided.

    Parameters
    ----------
    passbands : Passband or PassbandGroup, optional
        The passband or passband group to use for defining the light curve. If provided, they
        will be used to create box-shaped SED basis functions for each filter.
    filters : list, optional
        A list of filter names that the model supports. If None then
        all available filters will be used.
    """

    def __init__(self, *, passbands=None, filters=None, **kwargs):
        super().__init__(**kwargs)

        # Convert a single passband to a PassbandGroup.
        if isinstance(passbands, Passband):
            passbands = PassbandGroup(given_passbands=[passbands])

        # Create the SED basis functions for each filter.
        if passbands is not None:
            self.sed_basis = SEDBasisModel.from_box_approximation(passbands, filters=filters)
        else:
            self.sed_basis = None

        # Check that t0 is set.
        if "t0" not in kwargs or kwargs["t0"] is None:
            raise ValueError("Light curve models require a t0 parameter.")

    def compute_sed_given_lc(self, lc, times, wavelengths, graph_state):
        """Compute the flux density for a given light curve at specified times and wavelengths.

        Parameters
        ----------
        lc : LightcurveBandData
            The light curve data to use for computing the flux density.
        times : numpy.ndarray
            A length T array of observer frame timestamps in MJD.
        wavelengths : numpy.ndarray, optional
            A length N array of observer frame wavelengths (in angstroms).
        graph_state : GraphState
            An object mapping graph parameters to their values.

        Returns
        -------
        flux_density : numpy.ndarray
            A length T x N matrix of observer frame SED values (in nJy).
        """
        if self.sed_basis is None:
            raise ValueError("SED basis functions are not defined for this model.")

        params = self.get_local_params(graph_state)

        # Shift the times for the model's t0 aligned with the light curve's reference epoch.
        shifted_times = times - params["t0"]

        flux_density = np.zeros((len(times), len(wavelengths)))
        for filter in lc.filters:
            # Compute the SED values for the wavelengths we are actually sampling.
            sed_waves = self.sed_basis.compute_sed(filter, wavelengths=wavelengths)

            # Compute the multipliers for the SEDs at different time steps along this light curve.
            # We use the light curve's baseline value for all times outside the light curve's range.
            sed_time_mult = lc.evaluate_bandfluxes(shifted_times, filter)

            # The contribution of this filter to the overall SED is the light curve's (interpolated)
            # value at each time multiplied by the SED values at each query wavelength.
            sed_flux = np.outer(sed_time_mult, sed_waves)
            flux_density += sed_flux

        # Return the total flux density from all light curves.
        return flux_density

    def plot_sed_basis(self, ax=None, figure=None):
        """Plot the basis functions for the SED.  This is a debugging
        function to help the user understand the SEDs produced by this
        model.

        Parameters
        ----------
        ax : matplotlib.pyplot.Axes or None, optional
            Axes, None by default.
        figure : matplotlib.pyplot.Figure or None
            Figure, None by default.
        """
        if self.sed_basis is None:
            raise ValueError("SED basis functions are not defined for this model.")
        self.sed_basis.plot(ax=ax, figure=figure)


class LightcurveTemplateModel(BaseLightcurveBandTemplateModel):
    """A model that generates either the SED or bandflux of a source based on
    given light curves in each band. When generating the bandflux, it interpolates
    the light curves directly. When generating the SED, the model uses a box-shaped SED
    for each filter such that the resulting flux density is equal to the light curve's
    value after passing through the passband filter.

    LightcurveTemplateModel supports both periodic and non-periodic light curves. If the
    light curve is not periodic then each light curve's given values will be interpolated
    during the time range of the light curve. Values outside the time range (before and
    after) will be set to the baseline value for that filter (0.0 by default).

    Periodic models require that each filter's light curve is sampled at the same times
    and that the value at the end of the light curve is equal to the value at the start
    of the light curve. The light curve epoch (lc_data_t0) is automatically set to the first
    time so that the t0 parameter corresponds to the shift in phase.

    The set of passbands used to configure the model MUST be the same as used
    to generate the SED (the wavelengths must match).

    Parameterized values include:
      * dec - The object's declination in degrees.
      * ra - The object's right ascension in degrees.
      * t0 - The t0 of the zero phase (if applicable), date.

    Notes
    -----
    If you are interested in generating SED-level data, use the SEDTemplateModel in
    src/lightcurvelynx/models/sed_template_model.py instead.

    Attributes
    ----------
    lightcurves : LightcurveBandData
        The data for the light curves, such as the times and bandfluxes in each filter.
    sed_values : dict
        A dictionary mapping filters to the SED basis values for that passband.
        These SED values are scaled by the light curve and added for the
        final SED.
    filters : list
        The list of filters in the light curves.

    Parameters
    ----------
    lightcurves : dict or numpy.ndarray
        The light curves can be passed as either:
        1) a LightcurveBandData instance,
        2) a dictionary mapping filter names to a (T, 2) array of the bandlfuxes in that filter
        where the first column is time and the second column is the flux density (in nJy), or
        3) a numpy array of shape (T, 3) array where the first column is time (in days), the
        second column is the bandflux (in nJy), and the third column is the filter.
    passbands : Passband or PassbandGroup or None
        The passband or passband group to use for defining the light curve. If provided (not None),
        these will be used to create box-shaped SED basis functions for each filter.
    lc_data_t0 : float
        The reference epoch of the input light curve. This is the time stamp of the input
        array that will correspond to t0 in the model. For periodic light curves, this either
        must be set to the first time of the light curve or set as 0.0 to automatically
        derive the lc_data_t0 from the light curve.
    periodic : bool
        Whether the light curve is periodic. If True, the model will assume that
        the light curve repeats every period.
        Default: False
    baseline : dict or None
        A dictionary of baseline bandfluxes for each filter. This is only used
        for non-periodic light curves when they are not active.
        Default: None
    """

    def __init__(
        self,
        lightcurves,
        passbands,
        lc_data_t0,
        *,
        periodic=False,
        baseline=None,
        **kwargs,
    ):
        # Store the light curve data, parsing out different formats if needed.
        if isinstance(lightcurves, LightcurveBandData):
            self.lightcurves = lightcurves
        else:
            self.lightcurves = LightcurveBandData(
                lightcurves,
                lc_data_t0,
                periodic=periodic,
                baseline=baseline,
            )
        self.filters = self.lightcurves.filters
        super().__init__(passbands=passbands, filters=self.filters, **kwargs)

        # Raise a warning if time extrapolation is provided but cannot be used.
        if "time_extrapolation" in kwargs and kwargs["time_extrapolation"] is not None:
            if periodic:
                logger.warning("time_extrapolation is provided, but is not used for periodic light curves. ")
            elif baseline is None:
                logger.warning(
                    "time_extrapolation is provided, but is not used for light curves without a baseline. "
                )

    def minphase(self, filter=None, **kwargs):
        """Get the minimum supported phase of the model (for this filter) in days.

        Parameters
        ----------
        filter : str
            The name of the filter (required). An error is raised if no value is provided.
        **kwargs : dict
            Additional keyword arguments, not used in this method.

        Returns
        -------
        minphase : float or None
            The minimum phase of the model (in days) or None
            if the model does not have a defined minimum phase.
        """
        if filter is None:
            raise ValueError("Filter must be provided to compute minphase.")

        if self.lightcurves.has_time_bounds:
            return self.lightcurves.min_times[filter]
        else:
            return None

    def maxphase(self, filter=None, **kwargs):
        """Get the maximum supported phase of the model (for this filter) in days.

        Parameters
        ----------
        filter : str
            The name of the filter (required). An error is raised if no value is provided.
        graph_state : GraphState, optional
            An object mapping graph parameters to their values. If provided,
            the function will use the graph state to compute the maximum wavelength.
        **kwargs : dict
            Additional keyword arguments, not used in this method.

        Returns
        -------
        maximum : float or None
            The maximum phase of the model (in days) or None
            if the model does not have a defined maximum phase.
        """
        if filter is None:
            raise ValueError("Filter must be provided to compute maxphase.")

        if self.lightcurves.has_time_bounds:
            return self.lightcurves.max_times[filter]
        else:
            return None

    def compute_sed(self, times, wavelengths, graph_state):
        """Draw effect-free observer frame flux densities.

        Parameters
        ----------
        times : numpy.ndarray
            A length T array of observer frame timestamps in MJD.
        wavelengths : numpy.ndarray, optional
            A length N array of observer frame wavelengths (in angstroms).
        graph_state : GraphState
            An object mapping graph parameters to their values.

        Returns
        -------
        flux_density : numpy.ndarray
            A length T x N matrix of observer frame SED values (in nJy). These are generated
            from non-overlapping box-shaped SED basis functions for each filter and
            scaled by the light curve values.
        """
        return self.compute_sed_given_lc(
            self.lightcurves,
            times,
            wavelengths,
            graph_state,
        )

    def compute_bandflux(self, times, filter, state, **kwargs):
        """Evaluate the model at the passband level for a single, given graph state.

        Parameters
        ----------
        times : numpy.ndarray
            A length T array of observer frame timestamps in MJD.
        filter : str
            The name of the filter.
        state : GraphState
            An object mapping graph parameters to their values with num_samples=1.
        **kwargs : dict
            Additional keyword arguments, not used in this method.

        Returns
        -------
        bandfluxes : numpy.ndarray
            A length T matrix of observer frame passband fluxes (in nJy).
        """
        params = self.get_local_params(state)

        # Check that the filters are all supported by the model.
        if filter not in self.lightcurves.lightcurves:
            raise ValueError(f"Filter '{filter}' is not supported by LightcurveTemplateModel.")

        # Shift the times for the model's t0 aligned with the light curve's reference epoch.
        shifted_times = times - params["t0"]
        bandfluxes = self.lightcurves.evaluate_bandfluxes(shifted_times, filter)
        return bandfluxes

    def plot_lightcurves(self, times=None, ax=None, figure=None):
        """Plot the underlying light curves. This is a debugging
        function to help the user understand the SEDs produced by this
        model.

        Parameters
        ----------
        times : numpy.ndarray or None, optional
            An array of timestamps at which to plot the light curves.
            If None, the function uses the timestamps from each light curve.
        ax : matplotlib.pyplot.Axes or None, optional
            Axes, None by default.
        figure : matplotlib.pyplot.Figure or None
            Figure, None by default.
        """
        self.lightcurves.plot_lightcurves(times=times, ax=ax, figure=figure)


class MultiLightcurveTemplateModel(BaseLightcurveBandTemplateModel):
    """A MultiLightcurveTemplateModel randomly selects a light curve at each evaluation
    computes the flux from that source. The models can generate either the SED or
    bandflux of a source based of given light curves in each band. When generating
    the bandflux, the model interpolates the light curves directly. When generating the SED,
    the model uses a box-shaped SED for each filter such that the resulting flux density
    is equal to the light curve's value after passing through the passband filter.

    MultiLightcurveTemplateModel supports both periodic and non-periodic light curves. If the
    light curve is not periodic then each light curve's given values will be interpolated
    during the time range of the light curve. Values outside the time range (before and
    after) will be set to the baseline value for that filter (0.0 by default).

    Periodic models require that each filter's light curve is sampled at the same times
    and that the value at the end of the light curve is equal to the value at the start
    of the light curve. The light curve epoch is automatically set to the first time
    so that the t0 parameter corresponds to the shift in phase.

    The set of passbands used to configure the model MUST be the same as used
    to generate the SED (the wavelengths must match).

    Parameterized values include:
      * dec - The object's declination in degrees.
      * ra - The object's right ascension in degrees.
      * t0 - The t0 of the zero phase (if applicable), date.

    Attributes
    ----------
    lightcurves : list of LightcurveBandData
        The data for each set of light curves.
    sed_values : dict
        A dictionary mapping filters to the SED basis values for that passband.
        These SED values are scaled by the light curve and added for the
        final SED.
    all_filters : set
        A set of all filters used by the light curves. This is the union of all
        filters used by each light curve in the lightcurves list.

    Parameters
    ----------
    lightcurves : list of LightcurveBandData
        The data for each set of light curves. One light curve will be randomly selected
        at each evaluation.
    passbands : Passband or PassbandGroup or None, optional
        The passband or passband group to use for defining the light curve. If provided (not None),
        these will be used to create box-shaped SED basis functions for each filter.
        Default: None
    weights : numpy.ndarray, optional
        A length N array indicating the relative weight from which to select
        a light curve at random. If None, all light curves will be weighted equally.
    """

    def __init__(
        self,
        lightcurves,
        passbands=None,
        *,
        weights=None,
        **kwargs,
    ):
        # Validate the light curve input and create a union of all filters used.
        all_filters = set()
        for lc in lightcurves:
            if not isinstance(lc, LightcurveBandData):
                raise TypeError("Each light curve must be an instance of LightcurveBandData.")
            all_filters.update(lc.filters)
        self.filters = list(all_filters)
        self.lightcurves = lightcurves

        super().__init__(passbands=passbands, filters=self.filters, **kwargs)

        all_inds = [i for i in range(len(lightcurves))]
        self._sampler_node = GivenValueSampler(all_inds, weights=weights)
        self.add_parameter(
            "selected_lightcurve",
            value=self._sampler_node,
            allow_gradient=False,
            description="Index of the light curve selected for sampling.",
        )

        # Assemble a list of baseline values for each filter across all light curves.
        # Create a parameter to track the baseline values for the selected light curve. The node
        # will automatically fill in the correct baseline value based on the index given by
        # the selected_lightcurve parameter.
        for fltr in self.filters:
            baselines = [lc.baseline.get(fltr, 0.0) for lc in lightcurves]
            baseline_selector = GivenValueSelector(baselines, self.selected_lightcurve)
            self.add_parameter(
                f"baseline_{fltr}",
                value=baseline_selector,
                allow_gradient=False,
                description=f"Baseline value for filter {fltr} from the selected light curve.",
            )

    def __len__(self):
        """Get the number of light curves."""
        return len(self.lightcurves)

    @classmethod
    def from_lclib_file(cls, lightcurves_file, passbands, *, forced_lc_t0=None, filters=None, **kwargs):
        """Create a MultiLightcurveTemplateModel from a light curves file in LCLIB format.

        Parameters
        ----------
        lightcurves_file : str
            The path to the light curves file in LCLIB format.
        passbands : Passband or PassbandGroup
            The passband or passband group to use for defining the light curve.
        forced_lc_t0 : float or ndarray, optional
            By default we use the LCLIB convention of storing the light curves so the first time
            corresponds to the reference epoch (lc_data_t0) of the light curve. This can be overridden
            by providing a value for forced_lc_t0.
            Default: None
        filters : list of str, optional
            A list of filters to use for the light curves. If None, all filters will be used.
            Used to select a subset of filters that match the survey to simulate.
            Default: None
        **kwargs
            Additional keyword arguments to pass to the LightcurveBandData constructor, including
            the parameters for the model such as `dec`, `ra`, and `t0` and metadata
            such as `node_label`.

        Returns
        -------
        MultiLightcurveTemplateModel
            An instance of MultiLightcurveTemplateModel with the loaded light curves.
        """
        lightcurve_tables = read_lclib_data(lightcurves_file)
        if lightcurve_tables is None or len(lightcurve_tables) == 0:
            raise ValueError(f"Could not read light curves from file: {lightcurves_file}")

        if forced_lc_t0 is None:
            forced_lc_t0 = np.full(len(lightcurve_tables), None)
        elif np.isscalar(forced_lc_t0):
            forced_lc_t0 = np.full(len(lightcurve_tables), forced_lc_t0)
        elif len(forced_lc_t0) != len(lightcurve_tables):
            raise ValueError(
                "If provided as an array, forced_lc_t0 must have the same "
                "length as the number of light curves."
            )

        lightcurves = []
        for table, lc_t0 in tqdm(
            zip(lightcurve_tables, forced_lc_t0, strict=False), desc="Loading", unit="lc"
        ):
            lc_data = LightcurveBandData.from_lclib_table(table, forced_lc_t0=lc_t0, filters=filters)
            lightcurves.append(lc_data)

        # Add a citation for LCLIB if we loaded from an LCLIB file.
        cite_inline("LCLIB Data", f"LCLIB Data from the file {lightcurves_file}")

        return cls(lightcurves, passbands, **kwargs)

    def minphase(self, filter=None, graph_state=None, **kwargs):
        """Get the minimum supported phase of the model (for this filter) in days.

        Parameters
        ----------
        filter : str
            The name of the filter (required). An error is raised if no value is provided.
        graph_state : GraphState, optional
            An object mapping graph parameters to their values. If provided,
            the function will use the graph state to compute the minimum wavelength.
        **kwargs : dict
            Additional keyword arguments, not used in this method.

        Returns
        -------
        minphase : float or None
            The minimum phase of the model (in days) or None
            if the model does not have a defined minimum phase.
        """
        if filter is None:
            raise ValueError("Filter must be provided to compute minphase.")
        if graph_state is None:
            raise ValueError("Graph state must be provided to compute minphase.")
        if graph_state.num_samples > 1:
            raise ValueError("Graph state must have num_samples=1 to compute maxphase.")

        model_ind = self.get_param(graph_state, "selected_lightcurve")
        lc_model = self.lightcurves[model_ind]
        if lc_model.has_time_bounds:
            return lc_model.min_times[filter]
        else:
            return None

    def maxphase(self, filter=None, graph_state=None, **kwargs):
        """Get the maximum supported phase of the model (for this filter) in days.

        Parameters
        ----------
        filter : str
            The name of the filter (required). An error is raised if no value is provided.
        graph_state : GraphState, optional
            An object mapping graph parameters to their values. If provided,
            the function will use the graph state to compute the maximum wavelength.
        **kwargs : dict
            Additional keyword arguments, not used in this method.

        Returns
        -------
        maximum : float or None
            The maximum phase of the model (in days) or None
            if the model does not have a defined maximum phase.
        """
        if filter is None:
            raise ValueError("Filter must be provided to compute maxphase.")
        if graph_state is None:
            raise ValueError("Graph state must be provided to compute maxphase.")
        if graph_state.num_samples > 1:
            raise ValueError("Graph state must have num_samples=1 to compute maxphase.")

        model_ind = self.get_param(graph_state, "selected_lightcurve")
        lc_model = self.lightcurves[model_ind]
        if lc_model.has_time_bounds:
            return lc_model.max_times[filter]
        else:
            return None

    def compute_sed(self, times, wavelengths, graph_state):
        """Draw effect-free observer frame flux densities.

        Parameters
        ----------
        times : numpy.ndarray
            A length T array of observer frame timestamps in MJD.
        wavelengths : numpy.ndarray, optional
            A length N array of observer frame wavelengths (in angstroms).
        graph_state : GraphState
            An object mapping graph parameters to their values.

        Returns
        -------
        flux_density : numpy.ndarray
            A length T x N matrix of observer frame SED values (in nJy). These are generated
            from non-overlapping box-shaped SED basis functions for each filter and
            scaled by the light curve values.
        """
        # Use the light curve selected by the sampler node to compute the flux density.
        model_ind = self.get_param(graph_state, "selected_lightcurve")
        return self.compute_sed_given_lc(
            self.lightcurves[model_ind],
            times,
            wavelengths,
            graph_state,
        )

    def compute_bandflux(self, times, filter, state, **kwargs):
        """Evaluate the model at the passband level for a single, given graph state.

        Parameters
        ----------
        times : numpy.ndarray
            A length T array of observer frame timestamps in MJD.
        filter : str
            The name of the filter.
        state : GraphState
            An object mapping graph parameters to their values with num_samples=1.
        **kwargs : dict
            Additional keyword arguments, not used in this method.

        Returns
        -------
        bandfluxes : numpy.ndarray
            A length T matrix of observer frame passband fluxes (in nJy).
        """
        params = self.get_local_params(state)
        model_ind = params["selected_lightcurve"]
        lc = self.lightcurves[model_ind]

        # Check that the filter is supported by the model.
        if filter not in lc.lightcurves:
            raise ValueError(f"Filter '{filter}' is not supported by LightcurveTemplateModel {model_ind}.")

        # Shift the times for the model's t0 aligned with the light curve's reference epoch.
        shifted_times = times - params["t0"]
        bandfluxes = lc.evaluate_bandfluxes(shifted_times, filter)
        return bandfluxes



import numpy as np
import pytest
from citation_compass import find_in_citations
from lightcurvelynx.astro_utils.agn_utils import eddington_ratio_dist_fun


def test_eddington_ratio_dist_fun():
    """Test the eddington_ratio_dist_fun function."""
    # This is a change detection test to make sure the results match previous code.

    # Test that we can draw samples and the fall within the expected bounds.
    rng = np.random.default_rng(100)
    for type in ["blue", "red"]:
        for edd_ratio in [0.5, 1.0, 2.0]:
            samples = eddington_ratio_dist_fun(edd_ratio, type, rng, 1000)
            assert len(samples) == 1000
            assert np.all(samples >= 0.0)

    # Test that if we draw a single sample it is a float.
    sample = eddington_ratio_dist_fun(1.0, "blue")
    assert isinstance(sample, float)

    # Test that we fail if we use an incorrect galaxy type.
    with pytest.raises(ValueError):
        eddington_ratio_dist_fun(1.0, "green")


def test_eddington_ratio_dist_fun_citations():
    """Test the citations for the eddington_ratio_dist_fun function."""
    assert len(find_in_citations("Sartori et. al. 2019")) > 0
    assert len(find_in_citations("Weigel et. al. 2017")) > 0



from lightcurvelynx.models.physical_model import SEDModel


class SNIaHost(SEDModel):
    """A SN Ia host galaxy model with a hostmass parameter, more to be added.

    Parameterized values include:
      * dec - The object's declination in degrees. [from BasePhysicalModel]
      * distance - The object's luminosity distance in pc. [from BasePhysicalModel]
      * hostmass - The hostmass value in units of log10(M/M_solar).
      * ra - The object's right ascension in degrees. [from BasePhysicalModel]
      * redshift - The object's redshift. [from BasePhysicalModel]
      * t0 - The t0 of the zero phase, date. [from BasePhysicalModel]
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.add_parameter(
            "hostmass",
            description="The hostmass value in units of log10(M/M_solar).",
            **kwargs,
        )



#!/usr/bin/env python

import argparse
import sqlite3
from pathlib import Path
from shutil import copyfileobj

import fsspec


def make_opsim_shorten(opsim_path: str, n_rows: int, output_path: str):
    """Create a shortened version of an OpSim file.

    Parameters
    ----------
    opsim_path : `str`
        The fsspec-recognizable path to the OpSim SQLite file.
    n_rows : `int`
        The number of rows to keep in the shortened version.
    output_path : `str`
        The local path to save the shortened version.
    """
    # Save the whole opsim database to a new file
    with fsspec.open(opsim_path, mode="rb") as input_file:
        with open(output_path, "wb") as output_file:
            copyfileobj(input_file, output_file)

    table_to_halt = "observations"
    temp_table = f"temp_{table_to_halt}"
    with sqlite3.connect(output_path) as conn:
        cursor = conn.cursor()

        # Select and delete all tables but "observations"
        cursor.execute("SELECT name FROM sqlite_master WHERE type='table';")
        for (table,) in cursor.fetchall():
            if table != table_to_halt:
                cursor.execute(f"DROP TABLE {table};")

        # Create a temporary table to store the shortened version of "observations"
        cursor.execute(
            f"""
                   CREATE TABLE {temp_table} AS
                   SELECT * FROM {table_to_halt} LIMIT ?;
               """,
            (n_rows,),
        )
        # Drop the original table
        cursor.execute(f"DROP TABLE {table_to_halt};")
        # Rename the temporary table to the original table
        cursor.execute(f"ALTER TABLE {temp_table} RENAME TO {table_to_halt};")

        # Clean up
        cursor.execute("VACUUM;")

        conn.commit()


def parse_args(args):
    """Parse the command line arguments.

    Parameters
    ----------
    args : `list` of `str`
        The command line arguments.

    Returns
    -------
    `argparse.Namespace`
        The parsed arguments.
    """
    parser = argparse.ArgumentParser(description="Create a shortened version of an OpSim file.")
    parser.add_argument(
        "-i",
        "--input",
        default="http://astro-lsst-01.astro.washington.edu:8080/fbs_db/fbs_3.5/baseline/baseline_v3.5_10yrs.db",
        help="The fsspec-recognizable path to the OpSim SQLite file.",
    )
    parser.add_argument(
        "-n",
        "--n_rows",
        type=int,
        default=100,
        help="The number of rows to keep in the shortened version.",
    )
    default_output_path = Path(__file__).parent.parent / "data" / "opsim_shorten.db"
    parser.add_argument(
        "-o",
        "--output",
        default=str(default_output_path),
        help="The path to save the shortened version.",
    )
    return parser.parse_args(args)


def main(args=None):
    """Subsample an OpSim file and save it to a new file.

    Parameters
    ----------
    args : `list` of `str`, optional
        The command line arguments.
    """
    args = parse_args(args)
    make_opsim_shorten(opsim_path=args.input, n_rows=args.n_rows, output_path=args.output)


if __name__ == "__main__":
    main()



"""Flux-magnitude conversion utilities."""

import numpy as np
import numpy.typing as npt

# AB definition is zp=8.9 for 1 Jy
MAG_AB_ZP_NJY = 8.9 + 2.5 * 9


def mag2flux(mag: npt.ArrayLike) -> npt.ArrayLike:
    """Convert AB magnitude to bandflux in nJy

    Parameters
    ----------
    mag : ndarray of float
        The magnitude to convert to bandflux.

    Returns
    -------
    bandflux : ndarray of float
        The bandflux corresponding to the input magnitude.
    """
    return np.power(10.0, -0.4 * (mag - MAG_AB_ZP_NJY))


def flux2mag(flux_njy: npt.ArrayLike) -> npt.ArrayLike:
    """Convert bandflux in nJy to AB magnitude

    Parameters
    ----------
    flux_njy : ndarray of float
        The bandflux to convert to magnitude.

    Returns
    -------
    mag : ndarray of float
        The magnitude corresponding to the input bandflux.
    """
    return MAG_AB_ZP_NJY - 2.5 * np.log10(flux_njy)



"""The core functions for running the LightCurveLynx simulation."""

import concurrent.futures
import logging
from pathlib import Path

import numpy as np
import pandas as pd
from nested_pandas import NestedFrame
from tqdm import tqdm

from lightcurvelynx.astro_utils.noise_model import apply_noise
from lightcurvelynx.models.physical_model import BandfluxModel
from lightcurvelynx.utils.post_process_results import concat_results

logger = logging.getLogger(__name__)


class SimulationInfo:
    """A class to hold all the information (data and configuration) for a simulation
    run. This includes the model, the surveys, the passbands, etc. This object
    is used so we have have a single object to pass around to multiprocessing functions.

    Attributes
    ----------
    model : BasePhysicalModel
        The model to draw from. This may have its own parameters which
        will be randomly sampled with each draw.
    num_samples : int
        The number of samples.
    obstable : ObsTable or List of ObsTable
        The ObsTable(s) from which to extract information for the samples.
    passbands : PassbandGroup or List of PassbandGroup
        The passbands to use for generating the bandfluxes.
    apply_obs_mask: boolean
        If True, apply obs_mask to filter interesting indices/times.
    time_window_offset : tuple(float, float), optional
        A tuple specifying the time window offset (start, end) relative to t0 in days.
        This is used to filter the observations to only those within the specified
        time window (t0 + start, t0 + end). If None or the model does not have a
        t0 specified, no time window is applied.
    obstable_save_cols : list of str, optional
        A list of ObsTable columns to be saved as part of the results. This is used
        to save context information about how the light curves were generated. If the column
        is missing from one of the ObsTables, a null value such as None or NaN is used.
        If None, no additional columns are saved.
    param_cols : list of str, optional
        A list of the model's parameter columns to be saved as separate columns in
        the results (instead of just the full dictionary of parameters). These
        must be specified as strings in the node_name.param_name format.
        If None, no additional columns are saved.
    rng : numpy.random._generator.Generator, optional
        A given numpy random number generator to use for this computation. If not
        provided, the function uses the node's random number generator.
    sample_offset : int
        An offset to apply to the sample indices. This is used when splitting
        the simulation into multiple batches for multiprocessing.
    output_file_path : str or Path, optional
        The file path and name of where to save the results. If provided the results
        are saved to this file instead of being returned directly.
    kwargs : dict
        Additional keyword arguments to pass to the simulation function.
    """

    def __init__(
        self,
        model,
        num_samples,
        obstable,
        passbands,
        *,
        obstable_save_cols=None,
        param_cols=None,
        apply_obs_mask=False,
        time_window_offset=None,
        sample_offset=0,
        rng=None,
        output_file_path=None,
        **kwargs,
    ):
        self.model = model
        self.num_samples = num_samples
        self.obstable = obstable
        self.passbands = passbands
        self.apply_obs_mask = apply_obs_mask
        self.time_window_offset = time_window_offset
        self.obstable_save_cols = obstable_save_cols
        self.param_cols = param_cols
        self.sample_offset = sample_offset
        self.rng = rng
        self.kwargs = kwargs
        self.output_file_path = None

        if self.num_samples <= 0:
            raise ValueError("Number of samples must be a positive integer.")

        if output_file_path is not None:
            self.output_file_path = Path(output_file_path)
            if not self.output_file_path.parent.exists():
                raise ValueError(f"Output file directory {self.output_file_path.parent} does not exist.")

    def split(self, num_batches=None, batch_size=None):
        """Split the simulation info into multiple batches for parallel processing.

        Parameters
        ----------
        num_batches : int, optional
            The number of batches to split the simulation into. If None, batch_size
            must be provided.
        batch_size : int, optional
            The size of each batch. If None, num_batches must be provided.

        Returns
        -------
        batches : list of SimulationInfo
            A list of SimulationInfo objects, one for each batch.
        """
        if num_batches is None and batch_size is None:
            raise ValueError("Either num_batches or batch_size must be provided.")
        if num_batches is not None and batch_size is not None:
            raise ValueError("Either num_batches or batch_size must be provided, but not both.")

        if num_batches is not None:
            if num_batches <= 0:
                raise ValueError("num_batches must be a positive integer.")
            if num_batches > self.num_samples:
                num_batches = self.num_samples
            batch_size = np.ceil(self.num_samples / num_batches).astype(int)
        else:
            if batch_size <= 0:
                raise ValueError("batch_size must be a positive integer.")
            num_batches = np.ceil(self.num_samples / batch_size).astype(int)

        batches = []
        end_idx = 0
        for batch_idx in range(num_batches):
            # Compute the bounds of the batch.
            start_idx = end_idx
            end_idx = min(start_idx + batch_size, self.num_samples)
            batch_num_samples = end_idx - start_idx
            if batch_num_samples <= 0:
                break

            # If given a RNG, match sure we create a unique RNG for each batch.
            batch_rng = None
            if self.rng is not None:
                seed = self.rng.integers(0, 2**32 - 1)
                batch_rng = np.random.default_rng(seed)

            # If we are saving to a file, modify the output file path for this batch.
            if self.output_file_path is not None:
                batch_output_file_path = (
                    self.output_file_path.parent
                    / f"{self.output_file_path.stem}_part{batch_idx}{self.output_file_path.suffix}"
                )
            else:
                batch_output_file_path = None

            # Create a subset of the batch. Most information is the same (references to the
            # same objects), except for the number of samples and the RNG.
            batch_info = SimulationInfo(
                model=self.model,
                num_samples=batch_num_samples,
                obstable=self.obstable,
                passbands=self.passbands,
                obstable_save_cols=self.obstable_save_cols,
                param_cols=self.param_cols,
                apply_obs_mask=self.apply_obs_mask,
                time_window_offset=self.time_window_offset,
                sample_offset=self.sample_offset + start_idx,
                rng=batch_rng,
                output_file_path=batch_output_file_path,
                **self.kwargs,
            )
            batches.append(batch_info)
        return batches


def get_time_windows(t0, time_window_offset):
    """Get the time windows for each sample state based on the time window offset.

    Parameters
    ----------
    t0 : float or np.ndarray, optional
        The reference time (t0) for the time windows.
    time_window_offset : tuple(float, float), optional
        A tuple specifying the time window offset (start, end) relative to t0 in days.
        If None, no time window is applied.

    Returns
    -------
    start_times : np.ndarray or None
        The start times for each sample t0 + time_window_offset[0]. If a before time is given,
        this is always returned as an array (even if t0 is a scalar). None returned if there is
        no start time.
    end_times : np.ndarray or None
        The end times for each sample t0 + time_window_offset[1]. If an after time is given,
        this is always returned as an array (even if t0 is a scalar). None returned if there is
        no end time.
    """
    # If the model did not have a t0 or we do not have a time_window_offset,
    # we cannot apply a time window.
    if t0 is None or time_window_offset is None:
        return None, None
    if len(time_window_offset) != 2:
        raise ValueError("time_window_offset must be a tuple of (before, after) in days.")
    before, after = time_window_offset

    # If t0 is a scalar apply the offset directly.
    if np.isscalar(t0):
        t0 = np.array([t0])
    start_times = t0 + before if before is not None else None
    end_times = t0 + after if after is not None else None

    return start_times, end_times


def _simulate_lightcurves_batch(simulation_info):
    """Generate a number of simulations of the given model and information
    from one or more surveys.

    Parameters
    ----------
    simulation_info : SimulationInfo
        The information needed to perform simulate a single batch of results.

    Returns
    -------
    lightcurves : nested_pandas.NestedFrame or Path
        If no output_file_path is specified in simulation_info, a NestedFrame with a row
        for each object. Otherwise the NestedFrame is saved to a file and the function
        returns that file's path.
    """
    logger.info(
        f"Starting batch at {simulation_info.sample_offset} with {simulation_info.num_samples} samples."
    )

    # Extract the parameters from the SimulationInfo that are used repeated
    # (so we have shorter names).
    model = simulation_info.model
    num_samples = simulation_info.num_samples
    obstable = simulation_info.obstable
    passbands = simulation_info.passbands
    obstable_save_cols = simulation_info.obstable_save_cols
    rng = simulation_info.rng

    # Sample the parameter space of this model. We do this once for all surveys, so the
    # object use the same parameters across all observations.
    if num_samples <= 0:
        raise ValueError("Invalid number of samples.")
    logger.info(f"Sampling {num_samples} parameter sets from the model.")
    sample_states = model.sample_parameters(num_samples=num_samples, rng_info=rng)

    # If we are given information for a single survey, make it into a list.
    if not isinstance(obstable, list):
        obstable = [obstable]
    if not isinstance(passbands, list):
        passbands = [passbands]
    num_surveys = len(obstable)
    if num_surveys != len(passbands):
        raise ValueError("Number of surveys must match number of passbands.")

    # We do not currently support bandflux models with multiple surveys because
    # a bandflux model is defined relative to a single survey.
    if num_surveys > 1 and isinstance(model, BandfluxModel):
        raise ValueError(
            "Simulating a BandfluxModel with multiple surveys is currently not supported, "
            "because the bandflux model is defined relative to the filters of a single survey."
        )

    # Create a dictionary for the object level information, including any saved parameters.
    # Some of these are placeholders (e.g. nobs) until they can be filled in during the simulation.
    ra = np.atleast_1d(model.get_param(sample_states, "ra"))
    dec = np.atleast_1d(model.get_param(sample_states, "dec"))
    results_dict = {
        "id": [i for i in range(num_samples)],
        "ra": ra.tolist(),
        "dec": dec.tolist(),
        "nobs": [0] * num_samples,
        "t0": np.atleast_1d(model.get_param(sample_states, "t0")).tolist(),
        "z": np.atleast_1d(model.get_param(sample_states, "redshift")).tolist(),
        "params": [state.to_dict() for state in sample_states],
    }
    if simulation_info.param_cols is not None:
        for col in simulation_info.param_cols:
            if col not in sample_states:
                raise KeyError(
                    f"Parameter column {col} not found in model parameters. "
                    f"Available parameters are: {sample_states.get_all_params_names()}."
                )
            results_dict[col.replace(".", "_")] = np.atleast_1d(sample_states[col]).tolist()

    # Set up the nested array for the per-observation data, including ObsTable information.
    nested_index = []
    nested_dict = {
        "mjd": [],
        "filter": [],
        "flux": [],
        "fluxerr": [],
        "flux_perfect": [],
        "survey_idx": [],
        "is_saturated": [],
    }
    if obstable_save_cols is None:
        obstable_save_cols = []
    for col in obstable_save_cols:
        nested_dict[col] = []

    # Determine which of the of the simulated positions match the pointings from each ObsTable.
    logger.info("Performing range searches to find matching observations.")
    start_times, end_times = get_time_windows(
        model.get_param(sample_states, "t0"),
        simulation_info.time_window_offset,
    )
    all_obs_matches = [
        obstable[i].range_search(ra, dec, t_min=start_times, t_max=end_times) for i in range(num_surveys)
    ]

    # Get all times and all filters as numpy arrays so we can do easy subsets.
    all_times = [np.asarray(obstable[i]["time"].values, dtype=float) for i in range(num_surveys)]
    all_filters = [np.asarray(obstable[i]["filter"].values, dtype=str) for i in range(num_surveys)]

    # We loop over objects first, then surveys. This allows us to generate a single block
    # of data for the object over all surveys.
    logger.info("Simulating light curves for each object.")
    for idx, state in tqdm(enumerate(sample_states), total=num_samples, desc="Simulating", unit="obj"):
        total_num_obs = 0

        for survey_idx in range(num_surveys):
            # Find the indices and times where the current model is seen.
            obs_index = np.asarray(all_obs_matches[survey_idx][idx])
            if len(obs_index) == 0:
                obs_times = []
                obs_filters = []
            else:
                obs_times = all_times[survey_idx][obs_index]

                # Filter to only the "interesting" indices / times for this object.
                if simulation_info.apply_obs_mask:
                    obs_mask = model.mask_by_time(obs_times, state)
                    obs_index = obs_index[obs_mask]
                    obs_times = obs_times[obs_mask]
                # Extract the filters for this observation.
                obs_filters = all_filters[survey_idx][obs_index]

            # Compute the bandfluxes and errors over just the given filters.
            bandfluxes_perfect = model.evaluate_bandfluxes(
                passbands[survey_idx], obs_times, obs_filters, state
            )
            bandfluxes_error = obstable[survey_idx].bandflux_error_point_source(bandfluxes_perfect, obs_index)
            bandfluxes = apply_noise(bandfluxes_perfect, bandfluxes_error, rng=rng)

            # Apply saturation thresholds from the ObsTable.
            bandfluxes, bandfluxes_error, saturation_flags = obstable[survey_idx].compute_saturation(
                bandfluxes, bandfluxes_error, obs_index
            )

            # Append the per-observation data to the nested dictionary, including
            # any needed ObsTable columns.
            nobs = len(obs_times)
            nested_dict["mjd"].extend(list(obs_times))
            nested_dict["filter"].extend(list(obs_filters))
            nested_dict["flux_perfect"].extend(list(bandfluxes_perfect))
            nested_dict["flux"].extend(list(bandfluxes))
            nested_dict["fluxerr"].extend(list(bandfluxes_error))
            nested_dict["survey_idx"].extend([survey_idx] * nobs)
            nested_dict["is_saturated"].extend(list(saturation_flags))
            for col in obstable_save_cols:
                col_data = (
                    list(obstable[survey_idx][col].values[obs_index])
                    if col in obstable[survey_idx]
                    else [None] * nobs
                )
                nested_dict[col].extend(col_data)

            total_num_obs += nobs
            nested_index.extend([idx] * nobs)

        # The number of observations is the total across all surveys.
        results_dict["nobs"][idx] = total_num_obs

    # Create the nested frame and either save it to a file or return it directly.
    logger.info("Compiling results.")
    results = NestedFrame(data=results_dict, index=[i for i in range(num_samples)])
    nested_frame = pd.DataFrame(data=nested_dict, index=nested_index)
    results = results.add_nested(nested_frame, "lightcurve")
    if simulation_info.output_file_path is not None:
        results.to_parquet(simulation_info.output_file_path)
        return simulation_info.output_file_path
    return results


def _simulate_lightcurves_parallel(simulation_info, executor, batch_size=1_000):
    """Generate a number of simulations of the given model and information
    from one or more surveys.

    Parameters
    ----------
    simulation_info : SimulationInfo
        The information needed to perform simulate a single batch of results.
    executor : concurrent.futures.Executor, optional
        The executor object to use for parallel processing.
    batch_size : int, optional
        The number of samples to process in each batch when using multiprocessing.
        Default is 1000.

    Returns
    -------
    lightcurves : list of nested_pandas.NestedFrame or Path
        If no output_file_path is specified in simulation_info, a NestedFrame with a row
        for each object. Otherwise the NestedFrame is saved to a file and the function
        returns that file's path.
    """
    # Perform the simulation in parallel batches and combine the results. Since different
    # frameworks may return either Future objects or direct results from the map function,
    # we check for both.
    batches = simulation_info.split(batch_size=batch_size)
    futures_or_results = executor.map(_simulate_lightcurves_batch, batches)
    result_list = []
    for res in futures_or_results:
        if hasattr(res, "result") and callable(res.result):  # A Future
            result_list.append(res.result())
        else:  # A direct result
            result_list.append(res)
    return result_list


def simulate_lightcurves(
    model,
    num_samples,
    obstable,
    passbands,
    *,
    obstable_save_cols=None,
    param_cols=None,
    apply_obs_mask=False,
    time_window_offset=None,
    output_file_path=None,
    rng=None,
    executor=None,
    num_jobs=None,
    batch_size=100_000,
):
    """Generate a number of simulations of the given model and information
    from one or more surveys. The result data can either be returned directly
    (as a single nested data frame) or saved to file(s).

    Parameters
    ----------
    model : BasePhysicalModel
        The model to draw from. This may have its own parameters which
        will be randomly sampled with each draw.
    num_samples : int
        The number of samples.
    obstable : ObsTable or List of ObsTable
        The ObsTable(s) from which to extract information for the samples.
    passbands : PassbandGroup or List of PassbandGroup
        The passbands to use for generating the bandfluxes.
    apply_obs_mask: boolean
        If True, apply obs_mask to filter interesting indices/times.
    time_window_offset : tuple(float, float), optional
        A tuple specifying the time window offset (start, end) relative to t0 in days.
        This is used to filter the observations to only those within the specified
        time window (t0 + start, t0 + end). If None or the model does not have a
        t0 specified, no time window is applied.
    obstable_save_cols : list of str, optional
        A list of ObsTable columns to be saved as part of the results. This is used
        to save context information about how the light curves were generated. If the column
        is missing from one of the ObsTables, a null value such as None or NaN is used.
        If None, no additional columns are saved.
    param_cols : list of str, optional
        A list of the model's parameter columns to be saved as separate columns in
        the results (instead of just the full dictionary of parameters). These
        must be specified as strings in the node_name.param_name format.
        If None, no additional columns are saved.
    output_file_path : str or Path, optional
        The file path and name of where to save the results. If provided the results
        are saved to this file instead of being returned directly. If the simulation
        is run in parallel, multiple files are created with a _partN suffix.
    rng : numpy.random._generator.Generator, optional
        A given numpy random number generator to use for this computation. If not
        provided, the function uses the node's random number generator.
    executor : concurrent.futures.Executor, optional
        The executor object that to use for parallel processing. This can be any object that
        supports concurrent.futures's map() function, returning either Future objects or
        direct results. If None, the function runs in serial.
    num_jobs : int, optional
        If provided (and no executor is provided) creates a process pool (ProcessPoolExecutor)
        with the given number of workers to process the results.
    batch_size : int, optional
        The number of samples to process in each batch when using multiprocessing.
        Default is 100_000.

    Returns
    -------
    lightcurves : nested_pandas.NestedFrame or list of Path
        If output_file_path is None, a NestedFrame with a row for each object. Otherwise
        the file(s) are saved and the function returns a list of Paths to the saved files.
    """
    if output_file_path is not None:
        output_file_path = Path(output_file_path)
        output_file_path.parent.mkdir(parents=True, exist_ok=True)

    # Create the simulation info wrapper.
    simulation_info = SimulationInfo(
        model=model,
        num_samples=num_samples,
        obstable=obstable,
        passbands=passbands,
        rng=rng,
        apply_obs_mask=apply_obs_mask,
        time_window_offset=time_window_offset,
        obstable_save_cols=obstable_save_cols,
        param_cols=param_cols,
        output_file_path=output_file_path,
    )

    # If we do not have any parallelization information, perform in serial.
    if executor is None and num_jobs is None:
        return _simulate_lightcurves_batch(simulation_info)

    # If we have an executor, use that.
    if executor is not None:
        result_list = _simulate_lightcurves_parallel(simulation_info, executor, batch_size=batch_size)
    else:
        with concurrent.futures.ProcessPoolExecutor(max_workers=num_jobs) as executor:
            result_list = _simulate_lightcurves_parallel(simulation_info, executor, batch_size=batch_size)

    # If we returned the results directly, concatenate them into a single NestedFrame.
    # If we saved them to file, return the list of file paths.
    if output_file_path is None:
        results = concat_results(result_list)
    else:
        results = result_list
    return results


def compute_single_noise_free_lightcurve(
    model,
    graph_state,
    passbands,
    *,
    rest_frame_phase_min=-50.0,
    rest_frame_phase_max=50.0,
    rest_frame_phase_step=2.0,
):
    """Compute the noise-free light curve for a single object.

    This is a helper function for compute_noise_free_lightcurves.

    Parameters
    ----------
    model : BasePhysicalModel
        The model object to use for generating the light curves.
    graph_state : GraphState
        The state of the graph for the simulation. Must be a single state
        (num_samples=1).
    passbands : PassbandGroup
        The passbands to use for generating the bandfluxes.
    rest_frame_phase_min : float or np.ndarray
        The minimum rest-frame phase (in days) at which to evaluate the light curve.
        If an array is given, it must match the number of samples in graph_state.
        Default is -50.0 days.
    rest_frame_phase_max : float or np.ndarray
        The maximum rest-frame phase (in days) at which to evaluate the light curve.
        If an array is given, it must match the number of samples in graph_state.
        Default is 50.0 days.
    rest_frame_phase_step : float or np.ndarray
        The step size (in days) between rest-frame phases at which to evaluate the light curve.
        If an array is given, it must match the number of samples in graph_state.
        Default is 2.0 days.

    Returns
    -------
    lightcurves : dict
        A dictionary mapping each filter name to the corresponding array of bandfluxes (in nJy),
        with an additional keys "times" (in MJD) and "rest_phase" (in days relative to t0).
    """
    if graph_state.num_samples != 1:
        raise ValueError("graph_state must have num_samples=1.")

    # Generate the rest-frame times at which to evaluate the light curve.
    rest_phase = np.arange(rest_frame_phase_min, rest_frame_phase_max, step=rest_frame_phase_step)
    times = rest_phase.copy()

    # Compute the observed-frame times (MJD) if needed.
    redshift = model.get_param(graph_state, "redshift")
    if redshift is not None and redshift > 0:
        times = times * (1 + redshift)

    t0 = model.get_param(graph_state, "t0")
    if t0 is not None:
        times += t0

    # Compute the light curve without noise for each time in each filter.
    lightcurves = {"times": times, "rest_phase": rest_phase}
    for filter in passbands.filters:
        filters_array = np.full(len(times), filter)
        bandfluxes = model.evaluate_bandfluxes(passbands, times, filters_array, graph_state)
        lightcurves[filter] = bandfluxes
    return lightcurves


def compute_noise_free_lightcurves(
    model,
    graph_state,
    passbands,
    *,
    rest_frame_phase_min=-50.0,
    rest_frame_phase_max=50.0,
    rest_frame_phase_step=2.0,
):
    """Compute the noise-free light curves for a given model and one more more states
    at given times (in either MJD or rest-frame phase).

    This function simulates the light curves without adding any noise, allowing
    for the analysis of the underlying model behavior.

    Parameters
    ----------
    model : BasePhysicalModel
        The model object to use for generating the light curves.
    graph_state : GraphState
        The state of the graph for the simulation.
    passbands : PassbandGroup
        The passbands to use for generating the bandfluxes.
    rest_frame_phase_min : float or np.ndarray
        The minimum rest-frame phase (in days) at which to evaluate the light curve.
        If an array is given, it must match the number of samples in graph_state.
        Default is -50.0 days.
    rest_frame_phase_max : float or np.ndarray
        The maximum rest-frame phase (in days) at which to evaluate the light curve.
        If an array is given, it must match the number of samples in graph_state.
        Default is 50.0 days.
    rest_frame_phase_step : float or np.ndarray
        The step size (in days) between rest-frame phases at which to evaluate the light curve.
        If an array is given, it must match the number of samples in graph_state.
        Default is 2.0 days.

    Returns
    -------
    lightcurves : nested_pandas.NestedFrame
        A NestedFrame with a row for each object that contains a "lightcurve" column
        with a dictionary mapping each filter name to the corresponding array of
        bandfluxes (in nJy), with an additional nested columns "times" (in MJD) and
        "rest_phase" (in days relative to t0).
    """
    # Expand any scalar inputs to arrays.
    num_samples = graph_state.num_samples
    if np.isscalar(rest_frame_phase_min):
        rest_frame_phase_min = np.full(num_samples, rest_frame_phase_min)
    if np.isscalar(rest_frame_phase_max):
        rest_frame_phase_max = np.full(num_samples, rest_frame_phase_max)
    if np.isscalar(rest_frame_phase_step):
        rest_frame_phase_step = np.full(num_samples, rest_frame_phase_step)

    # Set up the per-model dictionary.
    results_dict = {
        "id": np.arange(num_samples).tolist(),
        "ra": np.atleast_1d(model.get_param(graph_state, "ra")).tolist(),
        "dec": np.atleast_1d(model.get_param(graph_state, "dec")).tolist(),
        "z": np.atleast_1d(model.get_param(graph_state, "redshift")).tolist(),
    }

    # Set up the nested dictionary with the light curve information.
    all_filters = passbands.filters
    nested_index = []
    nested_dict = {"times": [], "rest_phase": []}
    for filter in all_filters:
        nested_dict[filter] = []

    # Compute each light curve.
    for idx, state_i in enumerate(graph_state):
        lc = compute_single_noise_free_lightcurve(
            model,
            state_i,
            passbands,
            rest_frame_phase_min=rest_frame_phase_min[idx],
            rest_frame_phase_max=rest_frame_phase_max[idx],
            rest_frame_phase_step=rest_frame_phase_step[idx],
        )

        # Append the light curve data onto the nested dictionary.
        for key in nested_dict:
            nested_dict[key].extend(list(lc[key]))
        nested_index.extend([idx] * len(lc["times"]))

    # Create the nested results frame.
    results = NestedFrame(data=results_dict, index=[i for i in range(num_samples)])
    nested_frame = pd.DataFrame(data=nested_dict, index=nested_index)
    results = results.add_nested(nested_frame, "lightcurve")
    return results
